#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import os
from pyranges import read_gtf

from cptac.cancers.source import Source
import cptac.tools.dataframe_tools as df_tools
from cptac.cancers.mssm.mssm import Mssm

class WashuPdac(Source):
    def __init__(self, no_internet=False):
        """Define which dataframes as are available in the self.load_functions dictionary variable, with names as keys.

        Parameters:
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """
        
        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        self.data_files = {
            "cibersort"         : "CIBERSORT.Output_Abs_PDA.txt.gz",
            "CNV"               : "PDA.gene_level.from_seg.filtered.tsv.gz",
            "mapping"           : "gencode.v22.annotation.gtf.gz",
            "mature_miRNA"      : "PDAC_mature_miRNA_combined.tsv.gz",
            "precursor_miRNA"   : "PDAC_precursor_miRNA_combined.tsv.gz",
            "total_miRNA"       : "PDAC_total_miRNA_combined.tsv.gz",
            "miRNA"             : ["PDAC_mature_miRNA_combined.tsv.gz","PDAC_precursor_miRNA_combined.tsv.gz","PDAC_total_miRNA_combined.tsv.gz"],
            # "readme"            : ["README_miRNA","README_CIBERSORT", "README_xCell","README_somatic_mutation_WXS","README_gene_expression","README.boxnote","README_ESTIMATE_WashU"],
            "somatic_mutation"  : "PDA_discovery.dnp.annotated.exonic.maf.gz",
            "transcriptomics"   : ["PDA_NAT_RNA-Seq_Expr_WashU_FPKM.tsv.gz","PDA_tumor_RNA-Seq_Expr_WashU_FPKM.tsv.gz"],
            "tumor_purity"      : "CPTAC_pancan_RNA_tumor_purity_ESTIMATE_WashU.tsv.gz",
            "xcell"             : "PDA_xCell.txt.gz",
            "hla_typing": "hla.sample.ct.10152021.sort.tsv.gz"
        }

        #self._readme_files = {}

        self.load_functions = {
            'transcriptomics'   : self.load_transcriptomics,
            'somatic_mutation'  : self.load_somatic_mutation,
            'miRNA'             : self.load_miRNA,
            'xcell'             : self.load_xcell,
            'cibersort'         : self.load_cibersort,
            'CNV'               : self.load_CNV,
            'tumor_purity'      : self.load_tumor_purity,
            #'readme'            : self.load_readme,
            'hla_typing'          : self.load_hla_typing
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="pdac", source='washu', data_files=self.data_files, load_functions=self.load_functions, no_internet=no_internet)

    def load_transcriptomics(self):
        df_type = 'transcriptomics'
        # Check if transcriptomics data is not loaded yet
        if df_type not in self._data:
            # Locate the transcriptomics data files
            file_path_list = self.locate_files(df_type)
            
            # Loop over list of file paths
            for file_path in file_path_list:
                # Get a list of the levels of the path
                path_elements = file_path.split(os.sep)
                # Extract the file name
                file_name = os.path.basename(file_path)

                # If the file is tumor RNA-Seq expression data
                if file_name == "PDA_tumor_RNA-Seq_Expr_WashU_FPKM.tsv.gz":
                    # Load the data as a pandas DataFrame
                    df = pd.read_csv(file_path, sep='\t')
                    # Rename the columns to standard names
                    df = df.rename(columns={"gene_name": "Name","gene_id": "Database_ID"})
                    # Set "Name" and "Database_ID" as the index and sort
                    df = df.set_index(["Name", "Database_ID"]).sort_index()
                    # Transpose the DataFrame (samples as rows, genes as columns)
                    df = df.T
                    df.index.name = "Patient_ID"
                    # Remove the '-T' (for tumor) from patient IDs
                    df.index = df.index.str.replace(r"-T", "", regex=True)
                    # Save the DataFrame in helper_tables dict
                    self._helper_tables["transcriptomics_tumor"] = df

                # If the file is normal adjacent tissue RNA-Seq expression data
                if file_name == "PDA_NAT_RNA-Seq_Expr_WashU_FPKM.tsv.gz":
                    # Load the data as a pandas DataFrame
                    df_norm = pd.read_csv(file_path, sep='\t')
                    # Rename the columns to standard names
                    df_norm = df_norm.rename(columns={"gene_name": "Name","gene_id": "Database_ID"})
                    # Set "Name" and "Database_ID" as the index and sort
                    df_norm = df_norm.set_index(["Name", "Database_ID"]).sort_index()
                    # Transpose the DataFrame (samples as rows, genes as columns)
                    df_norm = df_norm.T
                    df_norm.index.name = "Patient_ID"
                    # Append '.N' (for normal) to patient IDs
                    df_norm.index = df_norm.index.str.replace(r"-A", ".N", regex=True)
                    # Save the DataFrame in helper_tables dict
                    self._helper_tables["transcriptomics_normal"] = df_norm

            # Retrieve tumor and normal DataFrames from helper_tables
            rna_tumor = self._helper_tables.get("transcriptomics_tumor")
            rna_normal = self._helper_tables.get("transcriptomics_normal")
            
            # Check if tumor and normal DataFrames exist and are pandas DataFrame objects
            if rna_tumor is None or rna_normal is None:
                print("rna_tumor or rna_normal is None")
                return
            if not isinstance(rna_tumor, pd.DataFrame) or not isinstance(rna_normal, pd.DataFrame):
                print("rna_tumor or rna_normal is not a DataFrame")
                return
    
            # Combine tumor and normal DataFrames vertically
            rna_combined = pd.concat([rna_tumor, rna_normal])

            # Save the combined DataFrame in self._data dictionary
            self.save_df(df_type, rna_combined)

    def load_somatic_mutation(self):
        df_type = 'somatic_mutation'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)

            df = pd.read_csv(file_path, sep='\t')    
            # Rename the columns we want to keep to the appropriate names
            df = pd.read_csv(file_path, sep='\t')    
            df['Patient_ID'] = df.loc[:, 'Tumor_Sample_Barcode']
            df = df.rename(columns={
                        "Hugo_Symbol":"Gene",
                        "Gene":"Gene_Database_ID",
                        "Variant_Classification":"Mutation",
                        "HGVSp_Short":"Location"})

            df = df.set_index("Patient_ID")
            df = df[ ['Gene'] + ["Mutation"] + ["Location"] + [ col for col in df.columns if col not in ["Gene","Mutation","Location"] ] ]
            df.index = df.index.str.replace(r"_T", "", regex=True)     
            # save df in self._data
            self.save_df(df_type, df)

    def load_miRNA(self):
        self.load_precursor_miRNA()
        self.load_mature_miRNA()
        self.load_total_mRNA()

    def load_precursor_miRNA(self):
        df_type = 'precursor_miRNA'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)

            df = pd.read_csv(file_path, delimiter = '\t', index_col = ['Name', 'ID','Alias'])
            df = df.transpose()
            df.index = df.index.str.replace('\.T$','', regex = True)
            df.index = df.index.str.replace('\.A$','.N', regex = True)
            df.index.name = 'Patient_ID'                
            # Sort
            normal = df.loc[df.index.str.contains('\.N$', regex =True)]
            normal = normal.sort_values(by=["Patient_ID"])
            tumor = df.loc[~ df.index.str.contains('\.N$', regex =True)]
            tumor = tumor.sort_values(by=["Patient_ID"])
            all_df = pd.concat([tumor, normal])
            # save df in self._data
            self.save_df('miRNA', all_df)

    def load_mature_miRNA(self):
        df_type = 'mature_miRNA'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)
            
            df = pd.read_csv(file_path, delimiter = '\t', index_col = ['Name', 'ID','Alias', 'Derives_from'])
            df = df.transpose()
            df.index = df.index.str.replace('\.T$','', regex = True)
            df.index = df.index.str.replace('\.A$','.N', regex = True)
            df.index.name = 'Patient_ID'                
            # Sort
            normal = df.loc[df.index.str.contains('\.N$', regex =True)]
            normal = normal.sort_values(by=["Patient_ID"])
            tumor = df.loc[~ df.index.str.contains('\.N$', regex =True)]
            tumor = tumor.sort_values(by=["Patient_ID"])
            all_df = pd.concat([tumor, normal])
            # save df in self._data
            self.save_df('miRNA', all_df)

    def load_total_mRNA(self):
        df_type = 'total_miRNA'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)
            
            df = pd.read_csv(file_path, delimiter = '\t', index_col = ['Name', 'ID','Alias'])
            df = df.transpose()
            df.index = df.index.str.replace('\.T$','', regex = True)
            df.index = df.index.str.replace('\.A$','.N', regex = True)
            df.index.name = 'Patient_ID'                
            # Sort
            normal = df.loc[df.index.str.contains('\.N$', regex =True)]
            normal = normal.sort_values(by=["Patient_ID"])
            tumor = df.loc[~ df.index.str.contains('\.N$', regex =True)]
            tumor = tumor.sort_values(by=["Patient_ID"])
            all_df = pd.concat([tumor, normal])
            # save df in self._data
            self.save_df('miRNA', all_df)

    def load_xcell(self):
        df_type = 'xcell'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)
        
            df = pd.read_csv(file_path, sep = '\t', index_col = 0) 
            df = df.transpose()
            df.columns.name = 'Name'
            df.index.name = 'Patient_ID'
            df.index = df.index.str.replace(r'-T$', '', regex=True) # remove label for tumor samples
            df.index = df.index.str.replace(r'-A$', '.N', regex=True) # change label for normal samples
            # save df in self._data
            self.save_df(df_type, df)

    def load_cibersort(self):
        df_type = 'cibersort'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)

            df = pd.read_csv(file_path, sep = '\t', index_col = 0) 
            df.index.name = 'Patient_ID'
            df.columns.name = 'Name'
            df.index = df.index.str.replace(r'-T$', '', regex=True) 
            df.index = df.index.str.replace(r'-A$', '.N', regex=True)
            # save df in self._data
            self.save_df(df_type, df)

    def load_mapping(self):
        df_type = 'mapping'
        if "CNV_gene_ids" not in self._helper_tables:
            file_path = self.locate_files(df_type)

            df = read_gtf(file_path)
            df = df.as_df()
            df = df[["gene_name","gene_id"]]
            df = df.drop_duplicates()
            df = df.rename(columns={"gene_name": "Name","gene_id": "Database_ID"})
            df = df.set_index("Name")
            self._helper_tables["CNV_gene_ids"] = df

    def load_CNV(self):
        df_type = 'CNV'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)

            df = pd.read_csv(file_path, sep="\t")
            df = df.rename(columns={"Gene": "Name"})
            df = df.set_index("Name")
            cnv = df

            self.load_mapping()
            gene_ids = self._helper_tables["CNV_gene_ids"]
            df = cnv.join(gene_ids,how = "left") #merge in gene_ids 
            df = df.reset_index()
            df = df.set_index(["Name", "Database_ID"]) #create multi-index
            df = df.T
            df.index.name = 'Patient_ID'
            # save df in self._data
            self.save_df(df_type, df)

    def load_tumor_purity(self):
        df_type = 'tumor_purity'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)
        
            df = pd.read_csv(file_path, sep = "\t", na_values = 'NA')
            df.Sample_ID = df.Sample_ID.str.replace(r'-T', '', regex=True) # only tumor samples in file
            df = df.set_index('Sample_ID') 
            df.index.name = 'Patient_ID'

            # get clinical df (used to slice out cancer specific patient_IDs in tumor_purity file)
            mssmclin = Mssm(filter_type='pdac', no_internet=self.no_internet)
            clinical_df = mssmclin.get_df('clinical')              
            patient_ids = clinical_df.index.to_list()
            df = df.loc[df.index.isin(patient_ids)]

            # save df in self._data
            self.save_df(df_type, df)

    def load_hla_typing(self):
        df_type = 'hla_typing'

        if df_type not in self._data:
            # perform initial checks and get file path (defined in source.py, the parent class)
            file_path = self.locate_files(df_type)

            # which cancer_type goes with which cancer in the mssm table
            tumor_codes = {'brca':'BR', 'ccrcc':'CCRCC',
                           'ucec':'UCEC', 'gbm':'GBM', 'hnscc':'HNSCC',
                           'lscc':'LSCC', 'luad':'LUAD', 'pdac':'PDA',
                           'hcc':'HCC', 'coad':'CO', 'ov':'OV'}

            df = pd.read_csv(file_path, sep='\t')
            df = df.loc[df['Cancer'] == tumor_codes[self.cancer_type]]
            df = df.set_index("Sample")
            df.index.name = 'Patient_ID'
            df = df.sort_values(by=["Patient_ID"])

            self.save_df(df_type, df)

        return self._data[df_type]


#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
from cptac.cancers.source import Source
from cptac import CPTAC_BASE_DIR

class UmichCoad(Source):
    def __init__(self, no_internet=False):
        """Define which dataframes as are available in the self.load_functions dictionary variable, with names as keys.

        Args:
            no_internet (bool, optional): If true, skips the index update step because it requires an internet connection.
        """

        self.data_files = {
            "proteomics" : "Report_abundance_groupby=protein_protNorm=MD_gu=2.tsv.gz",
            "phosphoproteomics" : "Report_abundance_groupby=multi-site_protNorm=MD_gu=2.tsv.gz",
            "mapping" : "CRC_Prospective sample info.xlsx",
        }
        
        self.load_functions = {
            'phosphoproteomics' : self.load_phosphoproteomics,
            'proteomics' : self.load_proteomics,
        }
        
        super().__init__(cancer_type="coad", source="umich", data_files=self.data_files, load_functions=self.load_functions, no_internet=no_internet)

    def load_mapping(self):
        """Loads mapping from 'Label' to 'Sample Code' into _helper_tables['map_ids'].

        Reads the 'CRC_Prospective sample info.xlsx' file and creates a dictionary mapping
        'Label' to 'Sample Code'. 'Label' refers to the unique identifiers assigned to the samples,
        while 'Sample Code' represents the patient IDs. This is used in subsequent load methods to map aliquots to patient IDs.
        """
        df_type = 'mapping'

        if not self._helper_tables:
            file_path = self.locate_files(df_type)
            
            df = pd.read_excel(file_path, index_col = 'Label', usecols = ['Label', 'Sample Code'])
            map_dict = df.to_dict()['Sample Code'] # Create dictionary with aliquots as keys and patient IDs as values
            self._helper_tables["map_ids"] = map_dict

    def load_phosphoproteomics(self):
        """Loads phosphoproteomics data into _data['phosphoprotemoics']
        
        Reads the 'Report_abundance_groupby=multi-site_protNorm=MD_gu=2.tsv.gz' file, processes it,
        and stores the result in _data['phosphoproteomics'].
        """
        df_type = 'phosphoproteomics'
        
        if df_type not in self._data:
            # perform initial checks and get file path (defined in source.py, the parent class)
            file_path = self.locate_files(df_type)
            
            # Load the dataset
            df = pd.read_csv(file_path, sep='\t') 

            # Splitting and processing the 'Index' column to get relevant information
            df[['Database_ID','Transcript_ID',"Gene_ID","Havana_gene","Havana_transcript","Transcript","Name","Site"]] = df.Index.str.split("\\|",expand=True)
            df[['num1','start',"end","detected_phos","localized_phos","Site"]] = df.Site.str.split("_",expand=True) 

            # Filtering out rows with unlocalized phosphorylation sites
            unlocalized_to_drop = df.index[~df["detected_phos"].eq(df["localized_phos"]) & \
                                           df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)]

            # Further filtering and processing
            df = df.drop(index=unlocalized_to_drop)

            df = df[df['Site'].notna()] # only keep columns with phospho site 
            df = df.set_index(['Name', 'Site', 'Peptide', 'Database_ID']) # create a multiindex in this order.
            #drop columns not needed in df 
            df.drop(['Gene',  "Index", "num1", "start", "end", "detected_phos", "localized_phos", "Havana_gene", "Havana_transcript", "MaxPepProb", "Gene_ID", "Transcript_ID", "Transcript"], axis=1, inplace=True)

            df = df.transpose() 
            ref_intensities = df.loc["ReferenceIntensity"]# Get reference intensities to use to calculate ratios 
            df = df.subtract(ref_intensities, axis="columns") # Subtract ref intensities from all the values, to get ratios
            df = df.iloc[1:,:] # drop ReferenceIntensity row 
            df.index.name = 'Patient_ID'
            
            # drop quality control and ref intensity cols
            drop_cols = ['colonRef22-2', 'RefInt_ColonRef01', 'RefInt_ColonRef02',
                   'RefInt_ColonRef03', 'RefInt_ColonRef04', 'RefInt_ColonRef05',
                   'RefInt_ColonRef06', 'RefInt_ColonRef07', 'RefInt_ColonRef08',
                   'RefInt_ColonRef09', 'RefInt_ColonRef10', 'RefInt_ColonRef11',
                   'RefInt_ColonRef12', 'RefInt_ColonRef13', 'RefInt_ColonRef14',
                   'RefInt_ColonRef15', 'RefInt_ColonRef16', 'RefInt_ColonRef17',
                   'RefInt_ColonRef18', 'RefInt_ColonRef19', 'RefInt_ColonRef20',
                   'RefInt_ColonRef21', 'RefInt_ColonRef22-1']
            df = df.drop(drop_cols, axis='index')
            
            # if self.version == "1.1":
            # FIXME: The following code was inside the if block. It should work fine without it.
            self.load_mapping()
            mapping_dict = self._helper_tables["map_ids"]

            df = df.reset_index()
            df['Patient_ID'] = df['Patient_ID'].replace(mapping_dict) # replace aliquots with Patient_IDs
            df.Patient_ID = df.Patient_ID.apply(lambda x: x[1:]+'.N' if x[0] == 'N' else x[1:]) # change normals to have .N
            df = df.set_index('Patient_ID')
            # /FIXME
            
            # save df in self._data
            self.save_df(df_type, df)

    def load_proteomics(self):
        """Loads proteomics data into _data['proteomics']
        
        Reads the 'Report_abundance_groupby=protein_protNorm=MD_gu.tsv.gz' file, processes it,
        and stores the result in _data['proteomics'].
        """
        df_type = 'proteomics'

        if df_type not in self._data:
            # perform initial checks and get file path (defined in source.py, the parent class)
            file_path = self.locate_files(df_type)

            df = pd.read_csv(file_path, sep='\t')
            df['Database_ID'] = df.Index.apply(lambda x: x.split('|')[0]) # get protein identifier 
            df['Name'] = df.Index.apply(lambda x: x.split('|')[6]) # get protein name 
            df = df.set_index(['Name', 'Database_ID']) # set multiindex
            df = df.drop(columns = ['Index', 'MaxPepProb', 'NumberPSM', 'Gene']) # drop unnecessary  columns
            df = df.transpose()
            ref_intensities = df.loc["ReferenceIntensity"] # get reference intensities to use to calculate ratios 
            df = df.subtract(ref_intensities, axis="columns") # subtract reference intensities from all the values
            df = df.iloc[1:,:] # drop ReferenceIntensity row 
            df.index.name = 'Patient_ID'

            # drop quality control and ref intensity cols
            drop_cols = ['colonRef22-2', 'RefInt_ColonRef01', 'RefInt_ColonRef02',
                   'RefInt_ColonRef03', 'RefInt_ColonRef04', 'RefInt_ColonRef05',
                   'RefInt_ColonRef06', 'RefInt_ColonRef07', 'RefInt_ColonRef08',
                   'RefInt_ColonRef09', 'RefInt_ColonRef10', 'RefInt_ColonRef11',
                   'RefInt_ColonRef12', 'RefInt_ColonRef13', 'RefInt_ColonRef14',
                   'RefInt_ColonRef15', 'RefInt_ColonRef16', 'RefInt_ColonRef17',
                   'RefInt_ColonRef18', 'RefInt_ColonRef19', 'RefInt_ColonRef20',
                   'RefInt_ColonRef21', 'RefInt_ColonRef22-1']
            df = df.drop(drop_cols, axis='index')

            self.load_mapping()
            mapping_dict = self._helper_tables["map_ids"]

            df = df.reset_index()
            df['Patient_ID'] = df['Patient_ID'].replace(mapping_dict) # replace aliquots with Patient_IDs
            df.Patient_ID = df.Patient_ID.apply(lambda x: x[1:]+'.N' if x[0] == 'N' else x[1:]) # change normals to have .N
            df = df.set_index('Patient_ID')
            # /FIXME

            # save df in self._data
            self.save_df(df_type, df)

    def load_acetylproteomics(self):
        """Loads acetylproteomics data into _data['acetylproteomics'].
        
        Reads the corresponding file, processes it,
        and stores the result in _data['acetylproteomics'].
        """
        df_type = 'acetylproteomics'

        if df_type not in self._data:
            # perform initial checks and get file path (defined in source.py, the parent class)
            file_path = self.locate_files(df_type)

            df = pd.read_csv(file_path, sep = "\t")
            # Parse a few columns out of the "Index" column that we'll need for our multiindex
            df[['Database_ID', "Site"]] = df.Index.str.split("_",expand=True)
            df = df[df['Site'].notna()] # only keep columns with phospho site

            # Load the gene names and merge them with the current dataframe based on 'Database_ID'
            df_gene_names = pd.read_csv(f"{CPTAC_BASE_DIR}/data/cptac_genes.csv")
            df_gene_names = df_gene_names.rename(columns={'Gene_Name': 'Name'}) # Renaming 'Gene_Name' to 'Name'
            df = pd.merge(df, df_gene_names, on='Database_ID', how='left')

            # Move 'Name' into the multiindex
            df = df.set_index(['Name', 'Site', 'Peptide', 'Database_ID']) # This will create a multiindex from these columns
            df = df.T # transpose
            ref_intensities = df.loc["ReferenceIntensity"]# Get reference intensities to use to calculate ratios
            df = df.iloc[1:,:] # drop ReferenceIntensity row

            # There was 1 duplicate ID (C3N-01825) in the proteomic and phosphoproteomic data.
            # I used the Payne lab mapping file "aliquot_to_patient_ID.tsv" to determine the tissue type
            # for these duplicates, and they were both tumor samples. Next, I ran a pearson correlation
            # to check how well the values from each duplicate correlated to its tumor flagship sample.
            # The first occurrence in the file had a higher correlation with the flagship sample
            # than the second occurrence. I also created scatterplots comparing each duplicate to its flagship sample.
            # We dropped the second occurrence of the duplicate because it didn't correlate very well to its flagship sample.

            # Get dictionary with aliquots as keys and patient IDs as values
            self.load_mapping()
            mapping_dict = self._helper_tables["map_ids"]
            df = df.rename(index = mapping_dict) # replace aliquots with patient IDs (normal samples have .N appended)
            # Add '.N' to enriched normal samples ('NX')
            df.index.name = 'Patient_ID'
            df = df.reset_index()
            df['Patient_ID'] = df['Patient_ID'].apply(lambda x: x+'.N' if 'NX' in x else x) # 'NX' are enriched normals
            df = df.set_index('Patient_ID')
            df = df_tools.rename_duplicate_labels(df, 'index') # add ".1" to the second ocurrence of the ID with a duplicate
            df = df.drop('C3N-01825.1', axis = 'index') # drop the duplicate that didn't correlate well with flagship

            # save df in self._data
            self.save_df(df_type, df)

#############################################

                
#       TODO: Readmes
#             elif file_name == "README_v3.boxnote":
#                 self._readme_files["readme_proteomics"] = get_boxnote_text(file_path)
                
#             elif file_name == "README.boxnote":
#                 self._readme_files["readme_phosphoproteomics"] = get_boxnote_text(file_path)

            '''
            if file_name == "S039_BCprospective_observed_0920.tsv.gz":
                df = pd.read_csv(file_path, sep='\t')
                df = df.transpose()
                df.index.name = 'Patient_ID'
                df.columns.name = 'Name'
                df = average_replicates(df)
                df = df.sort_values(by=["Patient_ID"])
                self._data["proteomics"] = df  
                
            if file_name == "S039_BCprospective_imputed_0920.tsv.gz":
                df = pd.read_csv(file_path, sep='\t')
                df = df.transpose()
                df.index.name = 'Patient_ID'
                df.columns.name = 'Name'
                df = average_replicates(df)
                df = df.sort_values(by=["Patient_ID"])
                self._data["proteomics_imputed"] = df
            '''


#   Copyright 2023 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Importing necessary libraries
import pandas as pd
import os
from cptac.cancers.source import Source

class BcmCoad(Source):
    def __init__(self, no_internet=False):
        """
        Define bcmcoad dataframes available in self.load_functions with names as keys.
        
        Parameters:
        no_internet (bool, optional): If set to True, skips the index update step which requires an internet connection.
        Default is False.
        """
        
        # Define the needed variables and pass them to the parent Dataset class __init__ function
        self.data_files = {
            "transcriptomics" : "CRC-gene_RSEM_tumor_normal_UQ_log2(x+1)_BCM.txt.gz",
            "mapping" : "gencode.v34.basic.annotation-mapping.txt.gz",
            "proteomics": ["COAD_proteomics_gene_abundance_log2_reference_intensity_normalized_Tumor.txt.gz", "COAD_proteomics_gene_abundance_log2_reference_intensity_normalized_Normal.txt.gz"],
            "phosphoproteomics" : ["COAD_phospho_site_abundance_log2_reference_intensity_normalized_Tumor.txt","COAD_phospho_site_abundance_log2_reference_intensity_normalized_Normal.txt.gz"],
            "CNV" : "COAD_WES_CNV_gene_ratio_log2.txt.gz",
            "miRNA" : "COAD_miRNAseq_mature_miRNA_RPM_log2_Tumor.txt.gz"
        }

        self.load_functions = {
            'transcriptomics' : self.load_transcriptomics,
            'proteomics': self.load_proteomics,
            'phosphoproteomics' : self.load_phosphoproteomics,
            'CNV' : self.load_CNV,
            'miRNA' : self.load_miRNA
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="coad", source='bcm', data_files=self.data_files, load_functions=self.load_functions, no_internet=no_internet)


    def load_mapping(self):
        """
        Load and parse all files for mapping. These will be used for transcriptomics loading.
        """
        df_type = 'mapping'
        
        # Check if helper tables have already been loaded
        if not self._helper_tables:
            file_path = self.locate_files(df_type)

            # Load the file, select needed columns, set index, remove duplicates
            df = pd.read_csv(file_path, sep="\t")
            df = df[["gene","gene_name"]] 
            df = df.set_index("gene")
            df = df.drop_duplicates()

            # Save mapping in helper tables
            self._helper_tables["gene_key"] = df


    def load_transcriptomics(self):
        """
        Load and parse all files for bcm coad transcriptomics data.
        """
        df_type = 'transcriptomics'
        
        # Check if data is already loaded
        if df_type not in self._data:
            # Get file path to the correct data
            file_path = self.locate_files(df_type)

            # Load the file 
            df = pd.read_csv(file_path, sep="\t")
            df.index.name = 'gene'

            # Load mapping information and add it to transcriptomics data
            self.load_mapping()
            gene_key = self._helper_tables["gene_key"]
            transcript = gene_key.join(df,how = "inner") 
            transcript = transcript.reset_index()
            transcript = transcript.rename(columns={"gene_name":"Name","gene":"Database_ID"})
            transcript = transcript.set_index(["Name", "Database_ID"])
            transcript = transcript.sort_index() 
            transcript = transcript.T
            transcript.index.name = "Patient_ID"

            df = transcript

            # Save df in data
            self.save_df(df_type, df)


    def load_proteomics(self):
        """
        Load and parse all files for bcm brca proteomics data
        """
        df_type = 'proteomics'

        # Check if data is already loaded
        if df_type not in self._data:
            # Get file path to the correct data
            file_path_list = self.locate_files(df_type)
            # Loop over list of file paths to load each type of proteomics data (tumor/normal)
            for file_path in file_path_list:
                file_name = os.path.basename(file_path)

                # Load and process the files
                if file_name == "COAD_proteomics_gene_abundance_log2_reference_intensity_normalized_Tumor.txt.gz":
                    df = pd.read_csv(file_path, sep='\t')
                    df.index.name = 'gene'

                    df.set_index('idx', inplace=True)

                    # Load mapping information
                    self.load_mapping()
                    gene_key = self._helper_tables["gene_key"]
                    # Join gene_key to df, reset index, rename columns, set new index and sort
                    tumor_proteomics = gene_key.join(df, how='inner')
                    tumor_proteomics = gene_key.join(df, how='inner')
                    tumor_proteomics = tumor_proteomics.reset_index()
                    tumor_proteomics = tumor_proteomics.rename(columns={"index": "Database_ID", "gene_name": "Name"})
                    tumor_proteomics = tumor_proteomics.set_index(["Name", "Database_ID"])
                    tumor_proteomics = tumor_proteomics.sort_index()  # alphabetize
                    tumor_proteomics = tumor_proteomics.T
                    tumor_proteomics.index.name = "Patient_ID"

                    df = tumor_proteomics

                    self._helper_tables["proteomics_tumor"] = df


                if file_name == "COAD_proteomics_gene_abundance_log2_reference_intensity_normalized_Normal.txt.gz":
                    df = pd.read_csv(file_path, sep='\t')
                    df.index.name = 'gene'

                    df.set_index('idx', inplace=True)
                    # Load mapping information
                    self.load_mapping()
                    gene_key = self._helper_tables["gene_key"]

                    # Join gene_key to df, reset index, rename columns, set new index and sort
                    normal_proteomics = gene_key.join(df, how='inner')
                    normal_proteomics = gene_key.join(df, how='inner')
                    normal_proteomics = normal_proteomics.reset_index()
                    normal_proteomics = normal_proteomics.rename(columns={"index": "Database_ID", "gene_name": "Name"})
                    normal_proteomics = normal_proteomics.set_index(["Name", "Database_ID"])
                    normal_proteomics = normal_proteomics.sort_index()  # alphabetize
                    normal_proteomics = normal_proteomics.T
                    normal_proteomics.index.name = "Patient_ID"
                    modified_index = [label + '.N' for label in normal_proteomics.index]
                    normal_proteomics.index = modified_index

                    df = normal_proteomics

                    self._helper_tables["proteomics_normal"] = df

            # Combine the two proteomics dataframes
            prot_tumor = self._helper_tables.get("proteomics_tumor")
            prot_normal = self._helper_tables.get("proteomics_normal") 
            prot_combined = pd.concat([prot_tumor, prot_normal])

            # Save df in data
            self.save_df(df_type, prot_combined)

    def load_phosphoproteomics(self):
        """
        Load and parse all files for bcm brca phosphoproteomics data
        """
        df_type = 'phosphoproteomics'

        # Check if data is already loaded
        if df_type not in self._data:
            # Get file path to the correct data
            file_path_list = self.locate_files(df_type)

            for file_path in file_path_list:

                file_name = os.path.basename(file_path)

                if file_name == "COAD_phospho_site_abundance_log2_reference_intensity_normalized_Tumor.txt":
                    # Load and process the file
                    df = pd.read_csv(file_path, sep='\t')
                    df.index.name = 'gene'

                    # Extract Database_ID, gene name, site, and peptide from 'idx' column
                    df[['Database_ID', 'Gene_Key', 'Site', 'Peptide']] = df['idx'].str.split('|', expand=True).iloc[:,
                                                                 [0, 1, 2, 3]]

                    # Load mapping information
                    self.load_mapping()
                    gene_key_df = self._helper_tables["gene_key"]
                    mapping = gene_key_df['gene_name'].to_dict()

                    # Map gene_key to get gene name
                    df['Name'] = df['Gene_Key'].map(mapping)

                    # Drop the 'idx' and 'Gene_Key' columns
                    df.drop(columns=['idx', 'Gene_Key'], inplace=True)

                    # Set the 'Name', 'Site', 'Peptide', and 'Database_ID' columns as index in this order
                    df.set_index(['Name', 'Site', 'Peptide', 'Database_ID'], inplace=True)

                    # Transpose the dataframe so that the patient IDs are the index
                    df = df.transpose()

                    # Rename the index to 'Patient_ID'
                    df.index.name = 'Patient_ID'

                    self._helper_tables["phosphoproteomics_tumor"] = df

                if file_name == "COAD_phospho_site_abundance_log2_reference_intensity_normalized_Normal.txt.gz":
                    # Load and process the file
                    df = pd.read_csv(file_path, sep='\t')
                    df.index.name = 'gene'

                    # Extract Database_ID, gene name, site, and peptide from 'idx' column
                    df[['Database_ID', 'Gene_Key', 'Site', 'Peptide']] = df['idx'].str.split('|', expand=True).iloc[:,
                                                                 [0, 1, 2, 3]]

                    # Load mapping information
                    self.load_mapping()
                    gene_key_df = self._helper_tables["gene_key"]
                    mapping = gene_key_df['gene_name'].to_dict()

                    # Map gene_key to get gene name
                    df['Name'] = df['Gene_Key'].map(mapping)

                    # Drop the 'idx' and 'Gene_Key' columns
                    df.drop(columns=['idx', 'Gene_Key'], inplace=True)

                    # Set the 'Name', 'Site', 'Peptide', and 'Database_ID' columns as index in this order
                    df.set_index(['Name', 'Site', 'Peptide', 'Database_ID'], inplace=True)

                    # Transpose the dataframe so that the patient IDs are the index
                    df = df.transpose()

                    # Rename the index to 'Patient_ID'
                    df.index.name = 'Patient_ID'
                    modified_index = [label + '.N' for label in df.index]
                    df.index = modified_index
                    self._helper_tables["phosphoproteomics_normal"] = df

            # Combine the two proteomics dataframes
            phospho_tumor = self._helper_tables.get("phosphoproteomics_tumor")
            phospho_normal = self._helper_tables.get("phosphoproteomics_normal")

            # Concatenate the two DataFrames
            phospho_combined = pd.concat([phospho_tumor, phospho_normal], axis=1)

            # Save df in data
            self.save_df(df_type, phospho_combined)

    def load_CNV(self):
        """Load and process the CNV data file, and save it in the _data dictionary."""
        df_type = 'CNV'
        # Check if data is already loaded
        if df_type not in self._data:
            # Get file path to the correct data
            file_path = self.locate_files(df_type)

            # Load and process the file
            df = pd.read_csv(file_path, sep='\t')
            df.index.name = 'gene'

            df.set_index('idx', inplace=True)
            # Load mapping information
            self.load_mapping()
            gene_key = self._helper_tables["gene_key"]

            # Join gene_key to df, reset index, rename columns, set new index and sort
            df = gene_key.join(df, how='inner')
            df = df.reset_index()
            df = df.rename(columns={"index": "Database_ID", "gene_name": "Name"})
            df = df.set_index(["Name", "Database_ID"])
            df = df.sort_index()  # alphabetize
            df = df.T
            df.index.name = "Patient_ID"

            # Save df in data
            self.save_df(df_type, df)

    def load_miRNA(self):
        """
        Load and parse all files for miRNA data
        """
        df_type = 'miRNA'

        # Check if data is already loaded
        if df_type not in self._data:
            # Get file path to the correct data
            file_path = self.locate_files(df_type)

            # Load and process the file
            df = pd.read_csv(file_path, sep='\t')

            # Here the idx will be the miRNA names and columns will be the patient IDs.
            df.set_index('idx', inplace=True)
            df.index.name = 'Name'  # Rename idx to Name
            df = df.transpose()  # Transpose the data frame to have miRNA as columns and patients as rows
            df.index.name = "Patient_ID"

            # Save df in data
            self.save_df(df_type, df)


import os
import requests
import concurrent.futures
from concurrent.futures import ThreadPoolExecutor
from hashlib import md5
from tqdm import tqdm
import cptac
from cptac.exceptions import *
import pandas as pd
import os.path as path

# Set directory constants
DATA_DIR = os.path.join(cptac.CPTAC_BASE_DIR, "data")
STATIC_DOI = '10.5281/zenodo.8394329'
RECORD_ID = STATIC_DOI.split('.')[-1]
ZENO_TOKEN = 'GijLB8joEFbeVEBQsjtJ8rH1uXMK8p5REgkNTfgHCMSR5LDyisZiZx1BRPQT'
AUTH_HEADER = {'Authorization': 'Bearer ' + ZENO_TOKEN}
BUCKET=None
INDEX_BUCKET = None
INDEX_DOI = '10.5281/zenodo.7897498'

def fetch_repo_data() -> dict:
    """Fetches the repo data from Zenodo, including metadata and file links."""
    repo_link = f"https://zenodo.org/api/records/{RECORD_ID}"
    response = requests.get(repo_link, headers=AUTH_HEADER)
    response.raise_for_status()
    return response.json()

def fetch_index_repo_data() -> dict:
    """Fetches the repo data from Zenodo, including metadata and file links."""
    repo_link = "https://zenodo.org/api/records/7897498"
    response = requests.get(repo_link, headers=AUTH_HEADER)
    response.raise_for_status()
    return response.json()



def init_files() -> None:
    """
    Initializes several files that are essential for cptac to run, such as the file index.
    Creates directory if it does not exist, fetches data from the repository, and writes
    index file.
    """
    os.makedirs(DATA_DIR, exist_ok=True)
    index_path = os.path.join(DATA_DIR, 'index.tsv')
    acetyl_mapping_path = os.path.join(DATA_DIR, 'cptac_genes.csv')
    brca_mapping_path = os.path.join(DATA_DIR, 'brca_mapping.csv')

    # Error handling for different exceptions
    try:
        repo_data = fetch_repo_data()
        index_repo_data = fetch_index_repo_data()
        global BUCKET
        global INDEX_BUCKET
        BUCKET = repo_data['files']
        INDEX_BUCKET = index_repo_data['files']
        index_data = []
        index_data.append(f"description\tfilename\tchecksum")
        for data_file in BUCKET:
            if data_file['key'].startswith('.'):
                continue # ignore hidden files
            filename_list = data_file['key'].split('-')
            description = '-'.join(filename_list[:3])
            index_data.append(f"{description}\t{'-'.join(filename_list)}\t{data_file['checksum'].split(':')[1]}")
        #with open(index_path, 'w') as index_file:
        #   index_file.write('\n'.join(index_data))
        #Download some other necessary files
        if not os.path.isfile(acetyl_mapping_path):
            for num in range(0, len(BUCKET) - 1):
                if BUCKET[num]['key'] == 'cptac_genes.csv':
                    get_data(BUCKET[num]['links']['self'], acetyl_mapping_path)
        if not os.path.isfile(brca_mapping_path):
            for num in range(0, len(BUCKET) - 1):
                if BUCKET[num]['key'] == 'brca_mapping.csv':
                    get_data(BUCKET[num]['links']['self'], brca_mapping_path)
        if not os.path.isfile(index_path):
            get_data("https://zenodo.org/api/records/10573663/files/index.tsv/content", index_path)
    except requests.ConnectionError:
        raise NoInternetError("Cannot initialize data files: No internet connection.")
    except requests.RequestException as e:
        raise HttpResponseError(f"Requesting data failed with the following error: {e}")
    except Exception as e:
        raise CptacError(f"ERROR: {e}")
    

def download(cancer: str, source: str, dtype: str, data_file: str) -> bool:
    """
    Downloads data files for a specific cancer, source, datatype, and file name from Zenodo

    :param cancer: The cancer type (e.g. 'brca').
    :param source: The data source (e.g. 'harmonized').
    :param dtype: The datatype of the files to download (e.g. 'clinical')
    :param data_file: The file name to download (look at index.tsv for examples).

    :return: True if data has successfully downloaded; raises error otherwise.
    """
    # Input validation
    if not cancer or not source or not dtype:
        raise InvalidParameterError("Cancer, source, and datatypes must be provided.")
    
    # Prepare for data download
    # The procedure varies depending on the source and type of data
    if source in ['harmonized', 'mssm']:
        description = f"{source}-all_cancers-{dtype}"
    elif source in ['washu'] and dtype in ['tumor_purity', 'hla_typing']:
        description = f"{source}-all_cancers-{dtype}"
    else:
        description = f"{source}-{cancer}-{dtype}"
    output_dir = '-'.join(description.split('-')[:-1])
    os.makedirs(os.path.join(DATA_DIR, output_dir), exist_ok=True)
    # Download requested dataframe
    file_name = f"{description}-{data_file}"
    output_file = os.path.join(output_dir, data_file)
    # Error handling for different exceptions
    try:
        repo_data = fetch_repo_data()
        for num in range(0, len(repo_data['files']) - 1):
                if repo_data['files'][num]['key'] == file_name:
                    get_data(repo_data['files'][num]['links']['self'], output_file)
        # Verify checksum
        with open(os.path.join(DATA_DIR, output_file), 'rb') as data_file:
            local_hash = md5(data_file.read()).hexdigest()
        if local_hash != cptac.INDEX.query("filename == @file_name")['checksum'].item():
            os.remove(os.path.join(DATA_DIR, output_file))
            raise DownloadFailedError("Download failed: local and remote files do not match. Please try again.")
        return True
    except NoInternetError as e:
        raise NoInternetError("Download failed -- No internet connection.")
    except DownloadFailedError as e:
        raise e
    except requests.RequestException as e:
        raise HttpResponseError(f"Requesting data failed with the following error: {e}")
    except Exception as e:
        raise DownloadFailedError(f"Failed to download data file for {source} {cancer} {dtype} with error:\n{e}") from e


def get_bucket() -> str:
    """
    Gets the bucket in Zenodo that houses all data files.
    :return: The URL of the Zenodo bucket containing the data files.
    """
    if BUCKET is None:
        init_files()
    return BUCKET


def download_chunk(url, start, end, file_path, pbar=None):
    """
    Downloads a chunk of a file and writes it to the specified file path.
    
    :param url: THE URL of the file to download.
    :param start: The start byte of the chunk to download.
    :param end: The end byte of the chunk to download.
    :param file_path: The path to the file where the downloaded chunk should be written
    """
    headers = {'Range': f'bytes={start}-{end}'}
    headers.update(AUTH_HEADER)
    response = requests.get(url, headers=headers, stream=True)
    response.raise_for_status()
    block_size = 1024 # We can adjust this value as needed

    with open(file_path, 'rb+') as data_file:
        data_file.seek(start)
        for data in response.iter_content(block_size):
            data_file.write(data)
            if pbar is not None:
                pbar.update(len(data)) # Update the progress bar based on the size of the data block


def get_data(url: str, subfolder: str = '', num_threads: int = 4) -> str:
    """
    Downloads a file using multithreading and saves it to the specified subfolder.

    :param url: The URL of the file to download.
    :param subfolder: The subfolder where the downloaded file should be saved.
    :param num_threads: The number of threads to use for downloading the file (default is 4).
    :return: The path of the downloaded file.
    """
    repo_data = fetch_repo_data()

    file_name = url.split('/')[-2]
    file_size = None # Initialize file_size
    for num in range(0, len(repo_data['files'])):
        if repo_data['files'][num]['key'] == file_name:
            file_size = repo_data['files'][num]['size']
            break
    
    if file_name == "index.tsv":
        file_size = 30203
        chunk_size = file_size // num_threads
    else:
        chunk_size = file_size // num_threads

    # Create an empty file with the same size as the file to be downloaded
    with open(os.path.join(DATA_DIR, subfolder), 'wb') as data_file:
        data_file.truncate(file_size)

    with ThreadPoolExecutor(max_workers=num_threads) as executor, tqdm(total=file_size, unit='B', unit_scale=True, desc=f"Downloading {os.path.split(subfolder)[1]}") as pbar:
        futures = []
        for i in range(num_threads):
            start = i * chunk_size
            end = start + chunk_size - 1 if i != num_threads - 1 else file_size - 1
            futures.append(executor.submit(download_chunk, url, start, end, os.path.join(DATA_DIR, subfolder), pbar))

        for future in concurrent.futures.as_completed(futures):
            future.result()  # Raise any exception encountered during download

    return subfolder




#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import scipy.stats
import statsmodels.stats.multitest
import operator
import warnings

from cptac.exceptions import InvalidParameterError, PvalWarning, StDevWarning

'''
@Param df:
    A dataframe containing the label column, and one or more real valued comparison columns.

@Param label_column:
    The name of the label column. This column must be in the dataframe, and must contain exactly 2 unique values.

@Param comparison_columns (default - will use all in dataframe):
    A list of columns on which t-tests will be performed. Each column must be in the dataframe, and must be real valued.
    If no value is specified, by default it will use every column in the dataframe, aside from the specified label column.

@Param alpha (default = .05):
    Significance level. Will be adjusted using parameter correction_method if more than 1 comparison is done.

@Param equal_var (default = True):
    Whether the variances of the two groups are equal. If True, will perform Student's t test. If False, will perform
    Welch's t test.

@Param return_all (default = False):
    Boolean. If true, will return a dataframe containing all comparisons and p-values, regardless of significance.
    If false, will only return significant comparisons and p-values in the dataframe, or None if no significant comparisons.

@Param correction_method (default = 'bonferroni'):
    String. Specifies method of adjustment for multiple testing. See -
    https://www.statsmodels.org/stable/generated/statsmodels.stats.multitest.multipletests.html
    - for documentation and available methods.

@Param mincount (default=3):
    The minimum number of samples that must have a recorded value, in order for a ttest to be performed. The default is 3,
    which means that there must be at least 3 values recorded for both the mutation type (e.g missense_mutation) and the wildtype.

@Param pval_return_corrected (default=True):
    Returns corrected pvalues if True,
    Returns uncorrected pvalues if False 

@Param quiet (default=False):
    Whether to silence cptac-generated warnings.

@Return:
    A pandas dataframe of column names and corresponding p-values which were determined to be significant in
    the comparison, sorted by significance (smallest p-values at the head). The 2 columns of the dataframe are
    'Comparison' and 'P_Value'.
    Returns None if dataframe was not formatted properly, or if no comparison was significant.

This method takes as a parameter a dataframe. Must be formatted in the following way. 1 column declared as the label column, with
the name of this column passed in as the second parameter. The Label column must contain exactly 2 unique entries,
and every row in the dataframe must have one of these 2 values in this column. The remaining columns will be real
valued columns on which t-tests will be done. A list of real valued columns on which to do t-tests will be passed in
as the third parameter. No t-test will be done on columns not included in this list.

The wrap_ttest method will then compare the two groups, as partitioned by the two values in the Label column, and
perform t-tests for each real valued column in the passed in list, generating a p-value.
The resulting p-values will be corrected for multiple testing, using a specified 'correction_method', and a dataframe with
the significant results will be returned as a dataframe, sorted by p-value.
'''

def wrap_ttest(df, label_column, comparison_columns=None, alpha=.05, equal_var=True, return_all=False, correction_method='bonferroni', mincount=3, pval_return_corrected=True, quiet=False):
    # Verify precondition that label column exists and has exactly 2 unique values
    label_values = df[label_column].unique()
    if len(label_values) != 2:
        raise ValueError("Incorrectly Formatted Dataframe! Label column must have exactly 2 unique values.")

    # Partition dataframe into two sets, one for each of the two unique values from the label column
    partition1 = df.loc[df[label_column] == label_values[0]]
    partition2 = df.loc[df[label_column] == label_values[1]]

    # If no comparison columns specified, use all columns except the specified labed column
    if not comparison_columns:
        comparison_columns = list(df.columns)
        comparison_columns.remove(label_column)

    # Determine the number of real valued columns on which we will do t-tests
    number_of_comparisons = len(comparison_columns)

    # Store comparisons and p-values in two arrays
    comparisons = []
    pvals = []

    # Loop through each comparison column, perform the t-test, and record the p-val
    for column in comparison_columns:
        if len(partition1[column].dropna(axis=0)) <= mincount:
            continue
        elif len(partition2[column].dropna(axis=0)) <= mincount:
            continue
        else:

            group1 = partition1[column].dropna(axis=0).astype(float)
            group2 = partition2[column].dropna(axis=0).astype(float)

            if (np.std(group1) == 0 or np.std(group2) == 0) and not quiet:
                warnings.warn(f"At least one group for column {column} had a standard deviation of zero.", StDevWarning)

            stat, pval = scipy.stats.ttest_ind(
                a=group1,
                b=group2, 
                equal_var=equal_var
            )

            if pd.notna(pval):
                comparisons.append(column)
                pvals.append(pval)
            elif not quiet:
                warnings.warn(f"pval for column {column} was NaN. pval dropped.", PvalWarning)

    if len(pvals) == 0: # None of the groups had enough members to pass the mincount
        raise InvalidParameterError("No groups had enough members to pass mincount; no tests run.")

    # Correct for multiple testing to determine if each comparison meets the new cutoff
    results = statsmodels.stats.multitest.multipletests(pvals=pvals, alpha=alpha, method=correction_method)
    reject = results[0]

    # Format results in a pandas dataframe
    results_df = pd.DataFrame(columns=['Comparison','P_Value'])

    # If return all, add all comparisons and p-values to dataframe
    if return_all:
        if pval_return_corrected:
            results_df['Comparison'] = comparisons
            results_df['P_Value'] = results[1]

        else:
            results_df['Comparison'] = comparisons
            results_df['P_Value'] = pvals

    # Else only add significant comparisons
    else:
        for i in range(0, len(reject)):
            if reject[i]:
                if pval_return_corrected:
                    new_result = pd.DataFrame({'Comparison':comparisons[i],'P_Value':results[1][i]}, index=[0])
                    results_df = pd.concat([results_df, new_result])
                else:
                    new_result = pd.DataFrame({'Comparison':comparisons[i],'P_Value':pvals[i]}, index=[0])
                    results_df = pd.concat([results_df, new_result])

    # Sort dataframe by ascending p-value
    results_df = results_df.sort_values(by='P_Value', ascending=True)
    results_df = results_df.reset_index(drop=True)

    return results_df

'''
@Param df: Dataframe.Each column is a different gene/ comparison. Rows contains numeric values (such as proteomics) for correlation test
@Param label_column: String. Name of column that will be your x axis and will be compared to all values in df unless otherwise specified.
@Param alpha: significant level
@Param comparison_columns: columns that will be looped through and used as y axis for correlation test.
All other columns beside label column unless specified here.
@Param correction_method: String. Specifies method of adjustment for multiple testing. See -
https://www.statsmodels.org/stable/generated/statsmodels.stats.multitest.multipletests.html
    - for documentation and available methods.
This function will return a data frame with the columns comparison, the correlation coefficient, and the p value.
'''
def wrap_pearson_corr(df,label_column, alpha=.05,comparison_columns=None,correction_method='bonferroni',return_all = True):


    #df = df.dropna(axis=1, how="all")

    '''If no comparison columns specified, use all columns except the specified labed column'''
    if not comparison_columns:
        comparison_columns = list(df.columns)
        comparison_columns.remove(label_column)
    '''Store comparisons,p-values, correlation in their own array'''
    comparisons = []
    pvals = []
    correlation=[]


    '''Format results in a pandas dataframe'''
    newdf = pd.DataFrame(columns=['Comparison','Correlation','P_value'])
    for gene in comparison_columns:
        #create subset df with interacting gene/ gene (otherwise drop NaN drops everything)
        df_subset = df[[label_column,gene]]
        #do a linear regression to see if it's a meaningful association
        #dropna will remove rows with nan
        df_subset = df_subset.dropna(axis=0, how="any")
        count_row = df_subset.shape[0]
        if count_row > 20:
            x1 = df_subset[[label_column]].values
            y1 = df_subset[[gene]].values
            x1 = x1[:,0]
            y1 = y1[:,0]
            corr, pval = scipy.stats.pearsonr(x1,y1)

            comparisons.append(gene)
            pvals.append(pval)
            correlation.append(corr)


    '''Correct for multiple testing to determine if each comparison meets the new cutoff'''
    results = statsmodels.stats.multitest.multipletests(pvals=pvals, alpha=alpha, method=correction_method)
    reject = results[0]

    if return_all:
        for i in range(0,len(comparisons)):
            newdf = newdf.append({'Comparison': comparisons[i],"Correlation": correlation[i],'P_value': pvals[i]}, ignore_index=True)

    '''Else only add significant comparisons'''
    if (return_all == False):
            for i in range(0, len(reject)):
                if reject[i]:
                    newdf = newdf.append({'Comparison': comparisons[i],"Correlation": correlation[i],'P_value': pvals[i]}, ignore_index=True)

    '''Sort dataframe by ascending p-value'''
    newdf = newdf.sort_values(by='P_value', ascending=True)
    '''If results df is not empty, return it, else return None'''
    return newdf

def permutation_test_means(group1, group2, num_permutations, paired=False):
    """Use permutation testing to calculate a P value for the difference between the means of two groups. You would use this instead of a Student's t-test if your data do not follow a normal distribution. Note that permutation tests are still subject to the assumption of the Student's t-test that if you want to see if the means of the two groups are different, they need to have the same variance.

    Parameters:
    group1 (pandas.Series): The first group of samples. NaNs will be dropped before any analysis. If doing a paired test, samples from the same subject must have the same index value in both group1 and group2.
    group2 (pandas.Series): The second group of samples. NaNs will be dropped before any analysis. If doing a paired test, samples from the same subject must have the same index value in both group1 and group2.
    num_permutations (int): The number of permutations to perform
    paired (bool, optional): Whether to do a paired test. Default is False.

    Returns:
    float: The difference between the means.
    float: The P value for the null hypothesis that the two groups have the same mean.
    """
    # Drop NaN values
    group1 = group1.dropna()
    group2 = group2.dropna()

    extreme_count = 0

    # Create an independent pseudo-random number generator
    generator = np.random.RandomState(0)

    if paired:

        # Calculate the paired differences, and the paired difference in the means
        paired_diffs = group1.combine(group2, operator.sub).dropna().values
        actual_diff = np.mean(paired_diffs)
        abs_actual_diff = abs(actual_diff)

        for i in range(num_permutations):

            # Randomly flip the signs of the differences and recalculate the mean
            random_signs = generator.choice([1, -1], size=paired_diffs.size)
            diffs_signs_perm = random_signs * paired_diffs
            perm_diff = np.mean(diffs_signs_perm)

            # Keep count of how many are as or more extreme than the actual difference
            if abs(perm_diff) >= abs_actual_diff: # We compare the absolute values for a two-tailed test
                extreme_count += 1

    else:
        # Concatenate the series
        both = group1.append(group2)

        # Calculate the actual difference in the means
        actual_diff = np.mean(group1) - np.mean(group2)
        abs_actual_diff = abs(actual_diff)

        # Set some local references to speed up lookups in the loop
        # We tried this optimization for other loops, but this is the only one where it made a difference
        generator_permutation = generator.permutation
        np_mean = np.mean
        both_values = both.values
        group1_size = group1.size

        for i in range(num_permutations):

            # Permute values
            perm_array = generator_permutation(both_values)

            # Split out permuted groups
            perm_group1 = perm_array[:group1_size]
            perm_group2 = perm_array[group1_size:]

            # Calculate the permutation's difference in the means
            perm_diff = np_mean(perm_group1) - np_mean(perm_group2)

            # Keep count of how many are as or more extreme than the actual difference
            if abs(perm_diff) >= abs_actual_diff: # We compare the absolute values for a two-tailed test
                extreme_count += 1

    # Calculate the P value
    P_val = extreme_count / num_permutations # Don't need to multiply by 2 because we compared the absolute values of difference between means.

    return actual_diff, P_val


def permutation_test_corr(data, num_permutations):
    """Use permutation testing to calculate a P value for the linear correlation coefficient between two variables in several samples. You would use this if your distribution didn't follow the Pearson correlation test's assumption of being bivariate normal.

    Parameters:
    data (pandas.DataFrame): A dataframe where the rows are samples, and the columns are the two variables we're testing correlation between.

    Returns:
    float: The linear correlation coefficient for the two variables.
    float: The P value for the null hypothesis that the correlation coefficient is zero.
    """

    # Check the table dimensions
    if data.shape[1] != 2:
        raise InvalidParameterError(f"Expected 2 columns in dataframe. Found {data.shape[1]}.")

    # Drop NaN values
    data = data.dropna()

    # Extract the values
    var1 = data.iloc[:, 0].values
    var2 = data.iloc[:, 1].values

    # Create an independent pseudo-random number generator
    generator = np.random.RandomState(0)

    # Calculate the actual correlation coefficient
    actual_coef = np.corrcoef(var1, var2)[0, 1]

    extreme_count = 0

    for i in range(num_permutations):
        var1_perm = generator.permutation(var1)
        perm_coef = np.corrcoef(var1_perm, var2)[0, 1]

        # Keep count of how many are as or more extreme than our coefficient
        if abs(perm_coef) >= abs(actual_coef): # We compare the absolute values for a two-tailed test
            extreme_count += 1

    # Calculate the P value
    P_val = extreme_count / num_permutations # Don't need to multiply by 2 because we compared the absolute values of coefficients.

    return actual_coef, P_val



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
from pyranges import read_gtf

from cptac.cancers.source import Source
from cptac.cancers.mssm.mssm import Mssm

class WashuBrca(Source):
    """The WashuBrca class is a subclass of the Source class in the CPTAC data package. It is responsible for loading and processing specific cancer data from the WashU source.

    Attributes:
        data_files (dict): A dictionary where the keys are names of data files, and the values are the corresponding file names.
        load_functions (dict): A dictionary where the keys are types of data, and the values are the corresponding data loading functions.
    
    """

    def __init__(self, no_internet=False):
        """Initialize the WashBrca class by setting up the data_files and load_functions dictionaries, and calling the parent class's __init__ function.

        Args:
            no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        self.data_files = {
            "cibersort"         : "CIBERSORT.Output_Abs_BR.txt.gz",
            "CNV"               : "BR.gene_level.from_seg.filtered.tsv.gz",
            "mapping"           : "gencode.v22.annotation.gtf.gz",
            # "readme"            : ["README_miRNA", "README_CIBERSORT", "README_xCell", "README_somatic_mutation_WXS", "README_gene_expression", "README.boxnote", "README_ESTIMATE_WashU"],
            "somatic_mutation"  : "BR_prospective.dnp.annotated.exonic.addrecovercases.maf.gz",
            "transcriptomics"   : "BR_tumor_RNA-Seq_Expr_WashU_FPKM.tsv.gz",
            "tumor_purity"      : "CPTAC_pancan_RNA_tumor_purity_ESTIMATE_WashU.tsv.gz",
            "xcell"             : "BR_xCell.txt.gz",
            #"not_used"         : #"BR_total_miRNA_combined.tsv",  no file on box yet
            "hla_typing": "hla.sample.ct.10152021.sort.tsv.gz"
        }

        #self._readme_files = {}
        
        self.load_functions = {
            'transcriptomics'   : self.load_transcriptomics,
            'somatic_mutation'  : self.load_somatic_mutation,
            'xcell'             : self.load_xcell,
            'cibersort'         : self.load_cibersort,
            'CNV'               : self.load_CNV,
            'tumor_purity'      : self.load_tumor_purity,
            #'readme'            : self.load_readme,
            "hla_typing": self.load_hla_typing
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="brca", source="washu", data_files=self.data_files, load_functions=self.load_functions, no_internet=no_internet)

    def load_transcriptomics(self):
        """Load and process the transcriptomics data file, and save it in the _data dictionary."""
        df_type = 'transcriptomics'
        if df_type not in self._data:
            # perform initial checks and get file path (defined in source.py, the parent class)
            file_path = self.locate_files(df_type)

            # process the file and add it to self._data
            df = pd.read_csv(file_path, sep="\t")
            df = df.rename(columns={"gene_name": "Name","gene_id": "Database_ID"})
            df = df.set_index(["Name", "Database_ID"])
            df = df.sort_index()
            df = df.T
            df.index.name = "Patient_ID"
            #remove label for tumor samples. All samples are tumors 
            df.index = df.index.str.replace(r"-T", "", regex=True) 
            # save df in self._data
            self.save_df(df_type, df)

    def load_somatic_mutation(self):
        """Load and process the somatic mutation data file, and save it in the _data dictionary."""
        df_type = 'somatic_mutation'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)

            df = pd.read_csv(file_path, sep='\t')    
            df['Patient_ID'] = df.loc[:, 'Tumor_Sample_Barcode']
            df = df.rename(columns={
                        "Hugo_Symbol":"Gene",
                        "Gene":"Gene_Database_ID",
                        "Variant_Classification":"Mutation",
                        "HGVSp_Short":"Location"})

            df = df.set_index("Patient_ID")
            df = df[ ['Gene'] + ["Mutation"] + ["Location"] + [ col for col in df.columns if col not in ["Gene","Mutation","Location"] ] ]
            df.index = df.index.str.replace(r"_T", "", regex=True)     
            # save df in self._data
            self.save_df(df_type, df)

    def load_xcell(self):
        """Load and process the xcell data file, and save it int the _data dictionary."""
        df_type = 'xcell'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)

            df = pd.read_csv(file_path, sep = '\t', index_col = 0) 
            df = df.transpose()
            df.columns.name = 'Name'
            df.index.name = 'Patient_ID'
            df.index = df.index.str.replace(r'-T$', '', regex=True) # remove label for tumor samples
            df.index = df.index.str.replace(r'-A$', '.N', regex=True) # change label for normal samples
            # save df in self._data
            self.save_df(df_type, df)

    def load_cibersort(self):
        """Load and process the cibersoty data file, and save it in the _data dictionary."""
        df_type = 'cibersort'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)
            
            df = pd.read_csv(file_path, sep = '\t', index_col = 0) 
            df.index.name = 'Patient_ID'
            df.columns.name = 'Name'
            df.index = df.index.str.replace(r'-T$', '', regex=True) 
            df.index = df.index.str.replace(r'-A$', '.N', regex=True)
            # save df in self._data
            self.save_df(df_type, df)

    def load_mapping(self):
        df_type = 'mapping'
        if "CNV_gene_ids" not in self._helper_tables:
            file_path = self.locate_files(df_type)

            df = read_gtf(file_path)
            df = df.as_df()
            df = df[["gene_name","gene_id"]]
            df = df.drop_duplicates()
            df = df.rename(columns={"gene_name": "Name","gene_id": "Database_ID"})
            df = df.set_index("Name")
            self._helper_tables["CNV_gene_ids"] = df

    def load_CNV(self):
        """Load and process the CNV data file, and save it in the _data dictionary."""
        df_type = 'CNV'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)

            df = pd.read_csv(file_path, sep="\t")
            df = df.rename(columns={"Gene": "Name"})
            df = df.set_index("Name")
            cnv = df

            self.load_mapping()
            gene_ids = self._helper_tables["CNV_gene_ids"]
            df = cnv.join(gene_ids,how = "left") #merge in gene_ids 
            df = df.reset_index()
            df = df.set_index(["Name", "Database_ID"]) #create multi-index
            df = df.T
            df.index.name = 'Patient_ID'
            # save df in self._data
            self.save_df(df_type, df)

    def load_tumor_purity(self):
        """Load and process the tumor purity data file, and save it in the _data dictionary."""
        df_type = 'tumor_purity'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)

            df = pd.read_csv(file_path, sep = "\t", na_values = 'NA')
            df.Sample_ID = df.Sample_ID.str.replace(r'-T', '', regex=True) # only tumor samples in file
            df = df.set_index('Sample_ID') 
            df.index.name = 'Patient_ID' 

            # Use mssm clinical to slice out cancer specific patient_IDs in tumor_purity file
            mssmclin = Mssm(filter_type='brca', no_internet=self.no_internet)
            clinical_df = mssmclin.get_df('clinical')
            patient_ids = clinical_df.index.to_list()
            df = df.loc[df.index.isin(patient_ids)]

            # save df in self._data
            self.save_df(df_type, df)

    def load_hla_typing(self):
        """Load and process the HLA typing data file, and save it in the _data dictionary."""
        df_type = 'hla_typing'

        if df_type not in self._data:
            # perform initial checks and get file path (defined in source.py, the parent class)
            file_path = self.locate_files(df_type)

            # which cancer_type goes with which cancer in the mssm table
            tumor_codes = {'brca':'BR', 'ccrcc':'CCRCC',
                           'ucec':'UCEC', 'gbm':'GBM', 'hnscc':'HNSCC',
                           'lscc':'LSCC', 'luad':'LUAD', 'pdac':'PDA',
                           'hcc':'HCC', 'coad':'CO', 'ov':'OV'}

            df = pd.read_csv(file_path, sep='\t')
            df = df.loc[df['Cancer'] == tumor_codes[self.cancer_type]]
            df = df.set_index("Sample")
            df.index.name = 'Patient_ID'
            df = df.sort_values(by=["Patient_ID"])

            self.save_df(df_type, df)

        return self._data[df_type]
    # TODO FIX
    # def load_readme(self):
    #     df_type = 'readme'
    #     if not self._readme_files: # if self._readme_files is empty
    #         file_path_list = self.perform_initial_checks(df_type)
    #         # loop over list of file paths
    #         for file_path in file_path_list:
    #             path_elements = file_path.split(os.sep) # Get a list of the levels of the path
    #             file_name = path_elements[-1]# The last element will be the name of the file. We'll use this to identify files for parsing in the if/elif statements below

    #             if file_name == "README_miRNA":
    #                 with open(file_path, 'r') as reader:
    #                     self._readme_files["readme_miRNA"] = reader.read()
                        
    #             elif file_name == "README_CIBERSORT":
    #                 with open(file_path, 'r') as reader:
    #                     self._readme_files["readme_cibersort"] = reader.read()
                        
    #             elif file_name == "README_xCell":
    #                 with open(file_path, 'r') as reader:
    #                     self._readme_files["readme_xcell"] = reader.read()
                
    #             elif file_name == "README_somatic_mutation_WXS":
    #                 with open(file_path, 'r') as reader:
    #                     self._readme_files["readme_somatic_mutation"] = reader.read()
                        
    #             elif file_name == "README_gene_expression":
    #                 with open(file_path, 'r') as reader:
    #                     self._readme_files["readme_transcriptomics"] = reader.read()
                
    #             elif file_name == "README.boxnote":
    #                 self._readme_files["readme_cnv"] = get_boxnote_text(file_path)
                    
    #             elif file_name == "README_ESTIMATE_WashU":
    #                 with open(file_path, 'r') as reader:
    #                     self._readme_files["readme_tumor_purity"] = reader.read()


import pytest
import cptac

def test_get_dataset():
    """Test that the dataset instantiation works as expected"""
    try:
        lung = cptac.Luad()
    except Exception as e:
        pytest.fail(f"Could not instantiate Luad dataset: {str(e)}")

@pytest.mark.parametrize("dataset", ['Luad', 'Brca', 'Ccrcc'])
def test_multiple_datasets(dataset):
    """Test that multiple datasets can be instantiated"""
    try:
        data = getattr(cptac, dataset)()
    except Exception as e:
        pytest.fail(f"Could not instantiate {dataset} dataset: {str(e)}")



#   Copyright 2023 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from cptac.cancers.cancer import Cancer

# Import various BRCA data sources
from cptac.cancers.bcm.bcmbrca import BcmBrca
from cptac.cancers.broad.broadbrca import BroadBrca
from cptac.cancers.umich.umichbrca import UmichBrca
from cptac.cancers.washu.washubrca import WashuBrca
from cptac.cancers.mssm.mssm import Mssm
from cptac.cancers.harmonized.harmonized import Harmonized

class Brca(Cancer):
    """Manages BRCA (Breast Cancer) data from various sources.
    
    This class extends the base Cancer class and initializes the BRCA data from 
    a variety of sources including BCM, Broad Institute, MSSM, University of Michigan, 
    Washington University, and a Harmonized dataset. 

    Attributes:
        _sources (dict): A dictionary holding data from different sources.
    """

    def __init__(self, no_internet=False):
        """Initializes the Brca object.
        
        Args:
            no_internet (bool): If True, the object will not attempt to download data from the internet. 
                                Default is False.

        Raises:
            ValueError: If the 'no_internet' argument is not of boolean type.
        """
        if not isinstance(no_internet, bool):
            raise ValueError("The 'no_internet' argument must be of boolean type.")

        super().__init__(cancer_type="brca")

        classes = {'bcm': BcmBrca, 'broad': BroadBrca, 'mssm': Mssm, 
                   'umich': UmichBrca, 'washu': WashuBrca, 'harmonized': Harmonized}
        
        # Populate the _sources attribute with data from different sources
        self._sources = {key: value(no_internet=no_internet) if key not in ['mssm', 'harmonized'] 
                         else value(filter_type='brca', no_internet=no_internet) 
                         for key, value in classes.items()}



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Imports from the cptac.cancers package
from cptac.cancers.cancer import Cancer

# Imports from the cptac.cancers sub-packages
from cptac.cancers.bcm.bcmccrcc import BcmCcrcc
from cptac.cancers.broad.broadccrcc import BroadCcrcc
from cptac.cancers.umich.umichccrcc import UmichCcrcc
from cptac.cancers.washu.washuccrcc import WashuCcrcc
from cptac.cancers.mssm.mssm import Mssm
from cptac.cancers.harmonized.harmonized import Harmonized


class Ccrcc(Cancer):
    """
    The Ccrcc class is a subclass of the Cancer class and provides an interface to
    various data sources with ccRCC (clear cell renal cell carcinoma) data.

    Attributes:
        _sources (dict): A dictionary that holds instances of data source objects.
    """

    def __init__(self, no_internet=False):
        """
        Constructor for the Ccrcc class. It initializes instances of various data source
        classes and adds them to the _sources attribute.

        Parameters:
            no_internet (bool): A flag indicating whether to download data or not. If True,
                                the data source objects are initialized without downloading
                                data. Defaults to False.
        """
        super().__init__(cancer_type="ccrcc")

        # Initialize data sources and add them to the _sources dictionary
        self._sources["bcm"] = BcmCcrcc(no_internet=no_internet)
        self._sources["broad"] = BroadCcrcc(no_internet=no_internet)
        self._sources["mssm"] = Mssm(filter_type='ccrcc', no_internet=no_internet)
        self._sources["umich"] = UmichCcrcc(no_internet=no_internet)
        self._sources["washu"] = WashuCcrcc(no_internet=no_internet)
        self._sources["harmonized"] = Harmonized(filter_type='ccrcc', no_internet=no_internet)



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import os
from pyranges import read_gtf
from cptac.cancers.source import Source

class BroadLscc(Source):
    def __init__(self, no_internet=False):
        """
        Initialize the BroadLscc class by setting up required parameters and
        calling the parent's __init__ function.

        Parameters:
        no_internet (bool, optional): If set to True, skips the index update step which requires an internet connection.
        This is useful in situations with spotty internet connections.
        """
        
        # Define data files and their corresponding loading functions
        self.data_files = {
            "transcriptomics" : "LSCC.rsem_transcripts_tpm.txt.gz",
            "mapping" : ["sample_descriptions.tsv.gz", "gencode.v34.GRCh38.genes.collapsed_only.gtf.gz"]
        }
        
        self.load_functions = {
            'transcriptomics' : self.load_transcriptomics,
        }
        
        # Call the parent class __init__ function
        super().__init__(cancer_type="lscc", source='broad', data_files=self.data_files, load_functions=self.load_functions, no_internet=no_internet)

    def load_mapping(self):
        """
        Load the mapping files and process them accordingly.

        This method locates the mapping files in the specified directory, reads the files,
        and processes the data to create a dictionary for broad keys and broad gene names.
        These dictionaries are then stored in the _helper_tables attribute for later use.
        """
        df_type = 'mapping'
        
        if not self._helper_tables:
            file_path_list = self.locate_files(df_type)
            for file_path in file_path_list:
                path_elements = file_path.split('/') # Split the path into components
                file_name = path_elements[-1] # Extract the file name
                
                # Load and process sample description
                if file_name == "sample_descriptions.tsv.gz":
                    broad_key = pd.read_csv(file_path, sep="\t")
                    broad_key = broad_key.loc[broad_key['cohort'] == "LSCC"] # Filter LSCC keys
                    broad_key = broad_key[["sample_id","GDC_id","tissue_type"]]
                    broad_key = broad_key.set_index("sample_id") # Set broad id as index
                    # Add tissue type to patient ID
                    broad_key["Patient_ID"] = broad_key["GDC_id"] + broad_key["tissue_type"] 
                    # Standardize patient ID format by removing "Tumor" and replacing "Normal" with ".N"
                    broad_key.Patient_ID = broad_key.Patient_ID.str.replace(r"Tumor", "", regex=True)
                    broad_key.Patient_ID = broad_key.Patient_ID.str.replace(r"Normal", ".N", regex=True)
                    # Convert dataframe to dictionary
                    broad_dict = broad_key.to_dict()["Patient_ID"]
                    self._helper_tables["broad_key"] = broad_dict
                    
                # Load and process gene names
                elif file_name == "gencode.v34.GRCh38.genes.collapsed_only.gtf.gz":
                    broad_gene_names = read_gtf(file_path)
                    broad_gene_names = broad_gene_names.as_df()
                    broad_gene_names = broad_gene_names[["gene_name","gene_id"]]
                    broad_gene_names = broad_gene_names.rename(columns= {"gene_name":"Name"}) # Change column name for merging
                    broad_gene_names = broad_gene_names.set_index("gene_id")
                    broad_gene_names = broad_gene_names.drop_duplicates()
                    self._helper_tables["broad_gene_names"] = broad_gene_names

    def load_transcriptomics(self):
        """
        Load transcriptomics data, process it and store it in the _data attribute.

        This method first checks if the transcriptomics data is already loaded.
        If not, it locates the transcriptomics file, reads the data, and processes it.
        It joins the data with gene names and renames the colmns with CPTAC IDs.
        The processed dataframe is then saved into the _data attribute.
        """
        df_type = 'transcriptomics'

        if df_type not in self._data:
            file_path = self.locate_files(df_type)
            
            df = pd.read_csv(file_path, sep="\t")
            df = df.set_index(["transcript_id","gene_id"])
            
            # Add gene names to transcriptomic data
            self.load_mapping()
            broad_gene_names = self._helper_tables["broad_gene_names"]
            broad_dict = self._helper_tables["broad_key"]        
            df = broad_gene_names.join(df, how = "left") # Merge in gene names, keep transcripts that have a gene name
            df = df.reset_index()
            df = df.rename(columns= {"transcript_id": "Transcript_ID","gene_id":"Database_ID"})
            df = df.set_index(["Name","Transcript_ID","Database_ID"])
            df = df.rename(columns = broad_dict) # Rename columns with CPTAC IDs
            df = df.sort_index() 
            df = df.T
            df.index.name = "Patient_ID"
            # Save dataframe in self._data
            self.save_df(df_type, df)



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   You may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from cptac.cancers.cancer import Cancer
from cptac.cancers.bcm.bcmucec import BcmUcec
from cptac.cancers.broad.broaducec import BroadUcec
from cptac.cancers.umich.umichucec import UmichUcec
from cptac.cancers.washu.washuucec import WashuUcec
from cptac.cancers.mssm.mssm import Mssm
from cptac.cancers.harmonized.harmonized import Harmonized

class Ucec(Cancer):
    """
    The Ucec class is a child class of the base Cancer class, intended for handling 
    Uterine Corpus Endometrial Carcinoma (UCEC) data from different sources.

    Attributes:
        _sources (dict): A dictionary of source objects that load and hold data 
                         pertaining to UCEC cancer.
    """

    def __init__(self, no_internet=False):
        """
        Initializes the Ucec object, loading the relevant data sources for UCEC data. 

        Parameters:
            no_internet (bool): If set to True, the data source objects are initialized 
                                without downloading data. Default is False.
        """
        super().__init__(cancer_type="ucec")

        # Initialize data source objects for various sources of UCEC data
        self._sources["bcm"] = BcmUcec(no_internet=no_internet)
        self._sources["broad"] = BroadUcec(no_internet=no_internet)
        self._sources["mssm"] = Mssm(filter_type='ucec', no_internet=no_internet)
        self._sources["umich"] = UmichUcec(no_internet=no_internet)
        self._sources["washu"] = WashuUcec(no_internet=no_internet)
        self._sources["harmonized"] = Harmonized(filter_type='ucec', no_internet=no_internet)




#   Copyright 2023 User Name
#   Licensed under the Apache License, Version 2.0 (the "License");
#   You may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
from cptac.cancers.source import Source

class Mssm(Source):
    TUMOR_CODES = {
        'brca':'BR', 'ccrcc':'CCRCC', 'ucec':'UCEC',
        'gbm':'GBM', 'hnscc':'HNSCC', 'lscc':'LSCC',
        'luad':'LUAD', 'pdac':'PDA', 'hcc':'HCC',
        'coad':'CO', 'ov':'OV'
    }
    def __init__(self, filter_type, no_internet=False):
        """Initialize Mssm object and load the available dataframes.

        Args:
        filter_type (str): The cancer type for which you want information. Mssm keeps all data in a single table, so to get data on a single cancer type all other types are filtered out.
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. Default is False.
        """
        self.data_files = {
            "clinical" : "clinical_Pan-cancer.May2022.tsv.gz",
        }

        self.load_functions = {
            'clinical' : self.load_clinical,
            'medical_history': self.load_medical_history,
            'follow-up': self.load_followup
        }

        super().__init__(cancer_type=filter_type, source='mssm', data_files=self.data_files, load_functions=self.load_functions, no_internet=no_internet)

    def _load_filtered_data(self, filter_keyword):
        """Load and filter the data based on a specific keyword.

        Args:
        filter_keyword (str): Keyword to filter the columns.

        Returns:
        DataFrame: The filtered data.
        """
        clinical_df = self.load_clinical()
        filtered_df = clinical_df[[col for col in clinical_df.columns if filter_keyword in col]]
        return filtered_df

    def load_clinical(self):
        """Load the clinical data and filters it based on filter_type.
        
        Returns:
        DataFrame: The filtered clinical data.
        """
        df_type = 'clinical'
        if df_type not in self._data:
            file_path = self.locate_files(df_type)
            tumor_codes = {'brca':'BR', 'ccrcc':'CCRCC',
                           'ucec':'UCEC', 'gbm':'GBM', 'hnscc':'HNSCC',
                           'lscc':'LSCC', 'luad':'LUAD', 'pdac':'PDA',
                           'hcc':'HCC', 'coad':'CO', 'ov':'OV'}

            df = pd.read_csv(file_path, sep='\t')
            df = df[df['tumor_code'] == tumor_codes[self.cancer_type]]
            df = df[df['discovery_study'] != 'No'] # Only keep discovery study = 'Yes' or 'na'
            df.set_index("case_id", inplace=True)
            df.index.name = 'Patient_ID'
            df.sort_values(by=["Patient_ID"], inplace=True)

            self.save_df(df_type, df)

        return self._data[df_type]

    def load_medical_history(self):
        """Load the medical history data by filtering the clinical data.
        
        Returns:
        DataFrame: The filtered medical history data.
        """
        df_type = 'medical_history'
        if df_type not in self._data:
            medical_history_df = self._load_filtered_data('medical_history')
            self.save_df(df_type, medical_history_df)

        return self._data[df_type]

    def load_followup(self):
        """Load the follow-up data by filtering the clinical data.
        
        Returns:
        DataFrame: The filtered follow-up data.
        """
        df_type = 'follow-up'
        if df_type not in self._data:
            followup_df = self._load_filtered_data('follow-up')
            self.save_df(df_type, followup_df)

        return self._data[df_type]


import pytest
import cptac
import pandas as pd

@pytest.mark.parametrize("dataset", ['Luad', 'Brca', 'Ccrcc'])
def test_consistency_across_calls(dataset):
    """Test for consistency across function calls"""
    data = getattr(cptac, dataset)()
    first_call = data.get_proteomics(source='umich')
    second_call = data.get_proteomics(source='umich')
    pd.testing.assert_frame_equal(first_call, second_call, check_dtype=True)



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import pandas as pd
import numpy as np
import json
import requests
import webbrowser
import warnings

from cptac.exceptions import HttpResponseError, InvalidParameterError, ParameterWarning

def get_interacting_proteins_string(protein, num_results=25):
    '''
    Retrieves a list of proteins known to interact with the specified protein from the STRING database.

    Parameters
    ----------
    protein: str
        The name of the protein to generate a list of interacting proteins for.
    num_results: int, optional
        The number of interacting proteins to retrieve. Defaults to 25.

    Returns
    -------
    pandas.Series
        A series of proteins known by the STRING API to interact with the specified protein. 
        This series always also contains the queried protein.

    Raises
    ------
    requests.exceptions.HTTPError
        If the STRING API request fails.
    '''

    # Send query to the STRING API
    query_url = "https://string-db.org/api/json/network"
    params = {
        "identifiers": protein,
        "species": "9606",  # Homo sapiens
        "limit": num_results,
    }

    response = requests.get(query_url, params=params)
    
    # Check that the response came back good
    response.raise_for_status()

    # Put the response data in a dataframe
    response_df = pd.DataFrame(response.json())

    # Get the unique values of interacting proteins
    interactors_A = response_df["preferredName_A"].unique()
    interactors_B = response_df["preferredName_B"].unique()

    # Combine and sort interactors
    all_interactors = np.sort(np.append(interactors_A, interactors_B))

    # Make sure the protein they searched is in the output list
    if protein not in all_interactors:
        all_interactors = np.insert(all_interactors, 0, protein)

    # Convert to series
    interactors_series = pd.Series(all_interactors)

    return interactors_series


def get_interacting_proteins_biogrid(protein):
    """
    Queries the BioGRID API to get interacting proteins for a given protein.

    Parameters
    ----------
    protein : str
        The name of the protein to generate a list of interacting proteins for.

    Returns
    -------
    pandas.DataFrame
        A dataframe of interacting proteins, ranked by the number of literature references supporting them.

    Raises
    ------
    requests.exceptions.HTTPError
        If the BioGRID API request fails.
    ValueError
        If no interactors are found for the specified protein.
    """

    # Post query to the BioGRID API
    query_url = "https://webservice.thebiogrid.org/interactions/"
    params = {
        "searchNames": "true",
        "geneList": protein,
        "includeInteractors": "true",
        "includeInteractorInteractions": "false",
        "interSpeciesExcluded": "true",
        "format": "json",
        "taxId": "9606",
        "start": "0",
        "accesskey": "0ff59dcf3511928e78aad499688381c9"
    }

    response = requests.get(query_url, params=params)
    
    # Check that the response came back good
    response.raise_for_status()

    # Raise error if no interactors found
    if len(response.json()) == 0:
        raise ValueError(f"No interactors found for '{protein}'. Are you sure you entered the identifier correctly?")

    # Put the response data in a dataframe
    response_df = pd.DataFrame(response.json()).transpose()

    # Create a series to store interactors and their counts
    interactor_counts = pd.Series(dtype=int)

    # Check each interactor, if it is not the protein, add to the count
    for interactor in response_df[["OFFICIAL_SYMBOL_A", "OFFICIAL_SYMBOL_B"]].values:
        interactor_counts[interactor[interactor != protein]] = interactor_counts.get(interactor[interactor != protein], 0) + 1

    # Convert the series to a dataframe and set column name
    interactors = interactor_counts.to_frame("num_references")

    # Set index name
    interactors.index.name = "protein"

    return interactors

def get_interacting_proteins_bioplex(protein, secondary_interactions=False):
    """
    Retrieves a list of interacting proteins for a given protein from the BioPlex data table.

    Parameters
    ----------
    protein : str
        The name of the protein to generate a list of interacting proteins for.
    secondary_interactions : bool, optional
        A flag to include secondary interactions. Default is False.

    Returns
    -------
    list or None
        A list of proteins which are interacting partners with the specified protein, 
        according to the BioPlex data table. Returns None if specified protein isn't 
        found, or no interacting partners are found.

    Raises
    ------
    FileNotFoundError
        If the BioPlex data file is not found.
    """

    file_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "data", "BioPlex_293T_Network_10K_Dec_2019.tsv.gz")

    # Raise error if file not found
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"BioPlex data file not found at {file_path}")

    # Load the BioPlex data
    bioplex_interactions = pd.read_csv(file_path, sep='\t')
    bioplex_interactions.sort_values(by=["pInt", "pNI"], ascending=[False, True], inplace=True)

    # Retrieve interacting proteins
    interactors = set(bioplex_interactions.loc[bioplex_interactions['SymbolA'] == protein, 'SymbolB'])
    interactors.update(bioplex_interactions.loc[bioplex_interactions['SymbolB'] == protein, 'SymbolA'])

    # Check for secondary interactions
    if secondary_interactions:
        secondary_interactors = set()
        for interactor in interactors:
            secondary = get_interacting_proteins_bioplex(interactor, False)
            if secondary is not None:
                secondary_interactors.update(secondary)

        interactors.update(secondary_interactors)

    # Check if there are any interactors
    if not interactors:
        return None

    return list(interactors)

    
def get_interacting_proteins_wikipathways(protein):
    """
    @param protein:
    	String. The name of the protein
    @Return:
    	A list of proteins known by the most recent WikiPathways download to be interacting parters with the specified protein.
    	Returns None if specified protein is not found in the WikiPathways dataframe (which was intersected with Uniprot).

    This function takes a path to WikiPathways Dataframe file and protein name and returns a list of all the proteins that interact with it, using the pathways from the WikiPathways relsease file.
    This function loads the WikiPathways dataframe, and iterates through the row labelled with that protein name, return every protein in a pathway that also contains that protein.
    """

    path_here = os.path.abspath(os.path.dirname(__file__))
    data_dir_name = "data"
    file_name = "WikiPathwaysDataframe.tsv.gz"
    file_path = os.path.join(path_here, data_dir_name, file_name)
    proteinName = protein

    df = pd.read_csv(file_path, sep="\t", index_col=0)

    if (proteinName in df.index):
        row = df.loc[proteinName]
        filtered_df = df.loc[:, row.values.tolist()]

        def has_true(values):
            for val in values:
                if val == True:
                    return True
            return False

        filtered_df_final = filtered_df.loc[filtered_df.apply(lambda row: has_true(row.values.tolist()), axis=1), :]
        return filtered_df_final.index.tolist()

    return list()  # The protein was not found.

def list_pathways_wikipathways():
    """
    @ Return:
    	A list of all the WikiPathways pathways

    Uses the WikipathwaysDataFrame to return a list of all the possible pathways found.
    """
    path_here = os.path.abspath(os.path.dirname(__file__))
    data_dir_name = "data"
    file_name = "WikiPathwaysDataframe.tsv.gz"
    file_path = os.path.join(path_here, data_dir_name, file_name)

    df = pd.read_csv(file_path, sep="\t", index_col=0)
    return list(df.columns)

def get_pathways_with_proteins(proteins, database, reactome_resource="UniProt", quiet=False):
    """
    Retrieves pathways containing a particular gene or protein from Reactome or WikiPathways.

    Parameters
    ----------
    proteins : str or list of str
        The protein(s) to look for matches to.
    database : str
        The database to use; either 'reactome' or 'wikipathways'.
    reactome_resource : str, optional
        If using Reactome, this is the resource the identifier(s) come from.
        Default is UniProt. Other options include HGNC, Ensembl, and GO.
    quiet : bool, optional
        Whether to suppress warnings issued when identifiers are not found. Default False.

    Returns
    -------
    pandas.DataFrame
        A table of pathways containing the given genes or proteins, with pathway names and
        their Reactome identifiers (if using Reactome).

    Raises
    ------
    ValueError
        If the database parameter is not 'reactome' or 'wikipathways'.
    """

    if isinstance(proteins, str):
        proteins = [proteins]

    if database.lower() == "reactome":
        all_pathway_df = get_pathways_from_reactome(proteins, reactome_resource, quiet)

    elif database.lower() == "wikipathways":
        all_pathway_df = get_pathways_from_wikipathways(proteins, quiet)

    else:
        raise ValueError("Invalid database. Please choose 'reactome' or 'wikipathways'.")

    return all_pathway_df.reset_index(drop=True)


def get_pathways_from_reactome(proteins, reactome_resource, quiet):
    headers = {"accept": "application/json"}
    params = {"species": "Homo sapiens"}
    all_pathway_df = pd.DataFrame()

    for id in proteins:
        url = f"https://reactome.org/ContentService/data/mapping/{reactome_resource}/{id}/pathways"
        resp = requests.get(url, headers=headers, params=params)

        # Error handling
        if resp.status_code != requests.codes.ok:
            if resp.status_code == 404:
                try:
                    msg = resp.json()["messages"]
                except (json.JSONDecodeError, KeyError):
                    raise HttpResponseError(f"Your query returned an HTTP status {resp.status_code}. The content returned from the request may be helpful:\n{resp.content.decode('utf-8')}") from None
                else:
                    if not quiet:
                        warnings.warn(f"The query for '{id}' returned HTTP 404 (not found). You may have mistyped the gene/protein ID or the reactome_resource name. The server gave the following message: {msg}", ParameterWarning, stacklevel=2)
                    continue
            else:
                raise HttpResponseError(f"Your query returned an HTTP status {resp.status_code}. The content returned from the request may be helpful:\n{resp.content.decode('utf-8')}")

        pathway_dict = resp.json()
        pathways = [{"id": id, "pathway": pathway["displayName"], "pathway_id": pathway["stId"]} for pathway in pathway_dict]

        all_pathway_df = all_pathway_df.append(pathways)

    return all_pathway_df.sort_values(by="pathway_id")


def get_pathways_from_wikipathways(proteins, quiet):
    file_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "data", "WikiPathwaysDataframe.tsv.gz")
    df = pd.read_csv(file_path, sep="\t", index_col=0)
    all_pathway_df = pd.DataFrame()

    for protein in proteins:
        if protein in df.index:
            pathways = df.columns[df.loc[protein, :]].values
            prot_df = [{"id": protein, "pathway": pathway} for pathway in pathways]
            all_pathway_df = all_pathway_df.append(prot_df)
        elif not quiet:
            warnings.warn(f"The protein '{protein}' was not found in the WikiPathways data.", ParameterWarning, stacklevel=2)

    return all_pathway_df

def get_proteins_in_pathways(pathways, database, quiet=False):
    """
    Retrieves a list of proteins contained in a particular pathway from Reactome or WikiPathways.

    Parameters
    ----------
    pathways : str or list of str
        The pathway(s) to get the contained proteins for.
        If using Reactome, these must be pathway IDs (e.g. "R-HSA-140877").
    database : str
        The database to use; either 'reactome' or 'wikipathways'.
    quiet : bool, optional
        Whether to suppress warnings issued when identifiers are not found. Default False.

    Returns
    -------
    pandas.DataFrame
        The proteins contained in the pathways.

    Raises
    ------
    ValueError
        If the database parameter is not 'reactome' or 'wikipathways'.
    """

    if isinstance(pathways, str):
        pathways = [pathways]

    if database.lower() == "reactome":
        all_protein_df = get_proteins_from_reactome(pathways, quiet)

    elif database.lower() == "wikipathways":
        all_protein_df = get_proteins_from_wikipathways(pathways, quiet)

    else:
        raise ValueError("Invalid database. Please choose 'reactome' or 'wikipathways'.")

    return all_protein_df.reset_index(drop=True)


def get_proteins_from_reactome(pathways, quiet):
    headers = {"accept": "application/json"}
    all_protein_df = pd.DataFrame()

    for pathway_id in pathways:
        url = f"https://reactome.org/ContentService/data/participants/{pathway_id}"
        resp = requests.get(url, headers=headers)

        # Error handling
        if resp.status_code != requests.codes.ok or (len(resp.content.decode("utf-8")) == 0 or len(resp.json()) == 0):
            if not quiet:
                warnings.warn(f"The query for '{pathway_id}' found no results. You may have mistyped the pathway ID.", ParameterWarning, stacklevel=2)
            continue

        members_df = pd.json_normalize(resp.json(), record_path=["refEntities"])
        prot_df = members_df[members_df["displayName"].str.startswith("UniProt:")]

        prot_names = prot_df["displayName"].str.rsplit(" ", n=1, expand=True)[1]
        proteins = [{"pathway": pathway_id, "member": name} for name in prot_names.unique()]

        all_protein_df = all_protein_df.append(proteins)

    return all_protein_df.drop_duplicates()


def get_proteins_from_wikipathways(pathways, quiet):
    file_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "data", "WikiPathwaysDataframe.tsv.gz")
    df = pd.read_csv(file_path, sep="\t", index_col=0)
    all_protein_df = pd.DataFrame()

    for pathway in pathways:
        if pathway in df.columns:
            prot_names = df.index[df[pathway]].values
            proteins = [{"pathway": pathway, "member": name} for name in prot_names]

            all_protein_df = all_protein_df.append(proteins)
        elif not quiet:
            warnings.warn(f"The pathway '{pathway}' was not found in the WikiPathways data.", ParameterWarning, stacklevel=2)

    return all_protein_df


def reactome_pathway_overlay(pathway, df=None, analysis_token=None, open_browser=True, export_path=None, image_format="png", display_col_idx=0, diagram_colors="Modern", overlay_colors="Standard", quality=7):
    """Visualize numerical data (e.g. protein expression) on a Reactome pathway diagram, with each node's color corresponding to the expression value provided for that molecule.

    Parameters:
    pathway (str): The Reactome ID for the pathway you want to overlay the data on, e.g. "R-HSA-73929".
    df (pandas.DataFrame or pandas.Series, optional): If you haven't previously analyzed your data with Reactome, give this parameter the data you want to overlay. Each row corresponds to a particular gene/protein/etc, and each column is expression or other data for a sample or aggregate. Index must be unique identifiers. Multiple data columns allowed. All dtypes must be numeric. Default None assumes you are instead passing a token for previously analyzed data to the "analysis_token" parameter.
    analysis_token (str, optional): If the data you want to visualize has been recently analyzed in Reactome already, pass the token for that analysis to this parameter to overlay it on the specified pathway. This will allow this function to reaccess the archived results, thus avoiding wasting time by repeating the work of submitting and analyzing the data. Default of None assumes you are instead passing data to the "df" parameter.
    open_browser (bool, optional): Whether to automatically open the diagram in a new web browser tab. Default True.
    export_path (str, optional): A string providing a path to export the diagram to. Must end in a file name with the same extension as the "image_format" parameter. Default None causes no figure to be exported.
    image_format (str, optional): If export_path is not none, this specifies the format to export the diagram to. Options are "png", "gif", "svg", "jpg", "jpeg", or "pptx". Must match the file extension in the export path. If you're creating a gif and you want more than one column's data to be included in the image, make sure to pass None to the display_col_idx parameter. Default "png".
    display_col_idx (int, optional): If export_path is not none, this specifies which column in the dataframe to overlay expression data from. Must be a valid column index for the given table, or None. None will cause the first column to be displayed, unless you're creating a gif, in which case it will cause all columns to be included in the gif. Default None.
    diagram_colors (str, optional): If export_path is not none, this specifies the Reactome color scheme to use for the underlying diagram. Options are "Modern" or "Standard". Default "Modern".
    overlay_colors (str, optional): If export_path is not none, this specifies the Reactome color scheme to use for the data overlay. Options are "Standard", "Strosobar", or "Copper Plus". Default "Standard".
    quality (int, optional): If export_path is not none, this specifies what relative quality to export the image at. Must be between 1 and 10 inclusive. Default 7.

    Returns:
    list of float: The mean of the data values for all proteins in the pathways, for each column in the data table, in the order of the columns in the data table. I.e. each value in this list is the average of the data from a particular column for all the proteins in the pathway.
    str: If export_path is None, returns URL to diagram with data overlaid in Reactome Pathway Browser. Otherwise returns the path the image was exported to.
    """
    # Parameter checking
    if df is None and analysis_token is None:
        raise InvalidParameterError("You passed None to both the 'df' and 'analysis_token' parameters. You must pass a value to one of them.")

    if df is not None and analysis_token is not None:
        raise InvalidParameterError("You passed values to both the 'df' and 'analysis_token' parameters. You may only pass a value to one of them.")

    if export_path is not None:

        if image_format not in ("png", "gif", "svg", "jpg", "jpeg", "pptx"):
            raise InvalidParameterError(f"Invalid value for 'image_format' parameter. Valid options are 'png', 'gif', 'svg', 'jpg', 'jpeg', or 'pptx'. You passed '{image_format}'.")

        if display_col_idx is None:
            display_col_idx = ""        
        elif df is not None and display_col_idx not in range(0, df.shape[1] if isinstance(df, pd.DataFrame) else 1):
            raise InvalidParameterError(f"Invalid value for 'display_col_idx' parameter. Must be either None, or an int between 0 and one less than the number of columns in df (which is {df.shape[1] - 1} for this df), inclusive. You passed {display_col_idx}.")

        if diagram_colors not in ("Modern", "Standard"):
            raise InvalidParameterError(f"Invalid value for 'diagram_colors' parameter. Valid options are 'Modern' or 'Standard'. You passed '{diagram_colors}'.")

        if overlay_colors not in ("Standard", "Strosobar", "Copper Plus"):
            raise InvalidParameterError(f"Invalid value for 'overlay_colors' parameter. Valid options are 'Standard', 'Strosobar', or 'Copper Plus'. You passed '{overlay_colors}'.")

        if quality not in range(1, 11):
            raise InvalidParameterError(f"Invalid value for 'quality' parameter. Must be an int between 1 and 10 inclusive. You passed {quality}.")

        if image_format != export_path.split('.')[-1]:
            raise InvalidParameterError(f"The file extension in the 'export_path' parameter must match the 'image_format' parameter. For the image_format parameter, you passed '{image_format}'. The extension at the end of your export path was '{export_path.split('.')[-1]}'.")

        if export_path[:2] == "~/":
            raise InvalidParameterError("The export path you provided appeared to start with a reference to the user home directory. To avoid confusion, this function will not expand that reference. Please provide a full path instead.")

    if df is not None:

        df = df.copy(deep=True)

        # If they gave us a series, make it a dataframe
        if isinstance(df, pd.Series):
            if df.name is None:
                df.name = "data"
            df = pd.DataFrame(df)

        # Check that the index is strings (gene/protein names)
        if df.index.dtype != np.dtype("object"):
            raise InvalidParameterError(f"The dataframe you passed does not have an index of strings. The dtype of your dataframe's index is {df.index.dtype}.")

        elif df.index.str.isnumeric().any():
            raise InvalidParameterError(f"The dataframe you passed has an index of strings, but some of the values are completely numbers as strings. Gene/protein identifier always have letters in them.")

        # The identifier series (the index) needs to have a name starting with "#"
        if df.index.name is None:
            df.index.name = "#identifier"
        elif not df.index.name.startswith("#"):
            df.index.name = "#" + df.index.name

        # Take care of NaNs
        df = df.astype(str) # This represents NaNs as 'nan', which Reactome is OK with

        # Get the df as a tab-separated string
        df_str = df.to_csv(sep='\t')

        # Post the data to the Reactome analysis service
        analysis_url = "https://reactome.org/AnalysisService/identifiers/projection"
        headers = {"Content-Type": "text/plain"}
        params = {"pageSize": "0", "page": "1"} # We only need the analysis token, so set pageSize to 0 so we don't worry about getting any of the data for individual pathways.

        view_resp = requests.post(analysis_url, headers=headers, params=params, data=df_str)

        # Check that the response came back good
        if view_resp.status_code != requests.codes.ok:
            raise HttpResponseError(f"Submitting your data for analysis returned an HTTP status {view_resp.status_code}. The content returned from the request may be helpful:\n{view_resp.content.decode('utf-8')}")    

        # Get the token for accessing the analysis results
        token = view_resp.json()["summary"]["token"]

    else:
        token = analysis_token

    # Get the mean data values
    expr_url = f"https://reactome.org/AnalysisService/token/{token}/filter/pathways?resource=TOTAL&pValue=1"

    headers = {
        "accept": "application/json",
        "content-type": "text/plain",
    }

    expr_resp = requests.post(expr_url, headers=headers, data=pathway)

    # Check that the response came back good
    if expr_resp.status_code != requests.codes.ok:
        raise HttpResponseError(f"Submitting your data for analysis returned an HTTP status {expr_resp.status_code}. The content returned from the request may be helpful:\n{expr_resp.content.decode('utf-8')}")    

    # Get the expression list
    expr_json = expr_resp.json()

    if len(expr_json) > 0:
        expr_list = expr_resp.json()[0]["entities"]["exp"]
    else:
        expr_list = []

    # Use the token and the pathway ID to open the pathway diagram with the data overlaid in the Reactome Pathway Browser
    viewer_url = f"https://reactome.org/PathwayBrowser/#/{pathway}&DTAB=AN&ANALYSIS={token}"
    if open_browser:
        webbrowser.open(viewer_url)

    if export_path is not None:

        # Get the diagram
        export_url = f"https://reactome.org/ContentService/exporter/diagram/{pathway}.{image_format}?token={token}&resource=TOTAL&diagramProfile={diagram_colors}&analysisProfile={overlay_colors}&expColumn={display_col_idx}&quality={quality}"
        export_resp = requests.get(export_url)

        # Check that the response came back good
        if export_resp.status_code != requests.codes.ok:
            raise HttpResponseError(f"Submitting your data for analysis returned an HTTP status {export_resp.status_code}. The content returned from the request may be helpful:\n{export_resp.content.decode('utf-8')}")    

        # Save the image
        with open(export_path, 'wb') as dest:
            dest.write(export_resp.content)

    if export_path is None:
        return expr_list, viewer_url
    else:
        return expr_list, export_path

def reactome_enrichment_analysis(analysis_type, data, sort_by, ascending, include_high_level_diagrams=True, disease_pathways=True, include_interactors=False):
    """Use the Reactome Analysis Service API to do a gene set enrichment analysis.

    Parameters:
    analysis_type (str): The type of enrichment analysis you want to perform. Either "ranked" or "nonranked".
    data (pandas.DataFrame or pandas.Series, or list or array-like): The data you want to overlay. Format depends on the analysis type.
        If ranked analysis:
            - data is a DataFrame or Series where the index is unique gene/protein identifiers and column(s) are ranking values (e.g. expression values for genes).
            - Multiple data columns allowed and are analyzed as separate ranked enrichment analyses.
            - All dtypes must be numeric.
        If unranked analysis:
            - data is a list or pandas.Index of identifiers to test pathways for enrichment with.
    sort_by (str): The metric by which to sort the pathways when selecting the top ones. You can pass "p_value" to sort by the P value (hypergeometric distribution), or pass one of the metrics directly supported by the Reactome API, listed below. (Yes, our function just maps "p_value" to "ENTITIES_PVALUE" .)
        "NAME",
        "TOTAL_ENTITIES",
        "TOTAL_INTERACTORS",
        "TOTAL_REACTIONS",
        "FOUND_ENTITIES",
        "FOUND_INTERACTORS",
        "FOUND_REACTIONS",
        "ENTITIES_RATIO", # Note: This value is the ratio of the total number of proteins in the pathway, to the total number of proteins in all of Reactome.
        "ENTITIES_PVALUE",
        "ENTITIES_FDR",
        "REACTIONS_RATIO",
    ascending (bool): When sorting pathways by the specified metric, whether to put smaller values first.
    include_high_level_diagrams (bool, optional): Whether to include pathway diagrams in the output that are just summaries of lower level pathways and don't show actual reactions. If False, this will exclude any Reactome pathways that have EHLD (Enhanced Higher Level Diagrams). Default True includes these pathways in results.
    disease_pathways (bool, optional): Whether to include pathways that describe disease related function. Default True.
    include_interactors (bool, optional): Whether to include computationally inferred interactors when identifying pathways that are enriched with your submitted proteins/genes. Default False. You may want to set this to True if a large portion of the identifiers you submitted do not match a Reactome pathway when it is set to False.

    Returns:
    pandas.DataFrame: A dataframe with info on the top enriched pathways, sorted by the specified metric.
    """
    # Check the sort_by parameter
    if sort_by == "p_value":
        parsed_sort_by = "ENTITIES_PVALUE"
    else:
        parsed_sort_by = sort_by

    valid_sort_bys = [
        "NAME",
        "TOTAL_ENTITIES",
        "TOTAL_INTERACTORS",
        "TOTAL_REACTIONS",
        "FOUND_ENTITIES",
        "FOUND_INTERACTORS",
        "FOUND_REACTIONS",
        "ENTITIES_RATIO",
        "ENTITIES_PVALUE",
        "ENTITIES_FDR",
        "REACTIONS_RATIO"]
    
    if parsed_sort_by not in valid_sort_bys: 
        newline = "\n"
        single_qt = "'"
        raise InvalidParameterError(f"Invalid value for 'sort_by' parameter. You passed '{sort_by}'. Must be one of the following:\n'p_value'\n{newline.join([f'{single_qt}{x}{single_qt}' for x in valid_sort_bys])}")

    if analysis_type == "ranked":

        # Copy the data
        data = data.copy(deep=True)
        
        # If they gave us a series, make it a dataframe
        if isinstance(data, pd.Series):
            if data.name is None:
                data.name = "data"
            data = pd.DataFrame(data)

        # Check that the index is strings (gene/protein names)
        if data.index.dtype != np.dtype("object"):
            raise InvalidParameterError(f"The dataframe you passed does not have an index of strings. The dtype of your dataframe's index is {data.index.dtype}.")

        elif data.index.str.isnumeric().any():
            raise InvalidParameterError(f"The dataframe you passed has an index of strings, but some of the values are completely numbers as strings. Gene/protein identifier always have letters in them.")

        # The identifier series (the index) needs to have a name starting with "#"
        if data.index.name is None:
            data.index.name = "#identifier"
        elif not data.index.name.startswith("#"):
            data.index.name = "#" + data.index.name

        # Take care of NaNs and small numbers
        # This represents NaNs as 'nan', which Reactome is OK with
        # Also rounds all numbers without scientific notation
        data.iloc[:, 0] = data.iloc[:, 0].apply(lambda x: "{:.9f}".format(x)) 

        # Get the data as a tab-separated string
        data_str = data.to_csv(sep='\t')

    elif analysis_type == "unranked":

        # Format data
        data = pd.Index(data) # Convert it to an index if it wasn't
        data = data.dropna() # Drop NaNs
        data = data.astype(str) # Make it strings

        # Check that they're actual gene/protein names
        if data.str.isnumeric().any():
            raise InvalidParameterError(f"The data you passed has some values that are completely numeric. Gene/protein identifier always have letters in them.")

        # The first item needs to be a column header string starting with '#'
        if not data[0].startswith("#"):
            data = data.insert(0, "#identifier")

        # Get the list as a newline-separated string
        data_str = "\n".join(data)

    else:
        raise InvalidParameterError(f"Invalid value for 'analysis_type' parameter. You passed '{analysis_type}'. Must be 'ranked' or 'unranked'.")

    # Post the data to the Reactome analysis service
    analysis_url = "https://reactome.org/AnalysisService/identifiers/projection"
    headers = {"Content-Type": "text/plain"}
    params = {
        "interactors": include_interactors,
        "sortBy": parsed_sort_by,
        "order": "ASC" if ascending else "DESC",
        "includeDisease": disease_pathways,
    }

    resp = requests.post(analysis_url, headers=headers, params=params, data=data_str)

    # Check that the response came back good
    if resp.status_code != requests.codes.ok:
        raise HttpResponseError(f"Submitting your data for analysis returned an HTTP status {resp.status_code}. The content returned from the request may be helpful:\n{resp.content.decode('utf-8')}")

    warnings_list = resp.json()["warnings"]
    if len(warnings_list) != 0:
        newline = "\n"
        raise InvalidParameterError(f"Your analysis request returned the following warnings. You may have a data formatting problem. Check that your data matches the format specified in the docstring. Here's up to the first ten warnings:\n{newline.join(warnings_list[0:len(warnings_list)] if len(warnings_list) < 10 else warnings_list[0:10] + ['...'])}")

    # Process the JSON response
    resp_json = resp.json()
    analysis_token = resp_json["summary"]["token"]
    pathways_table = pd.json_normalize(resp_json["pathways"], sep="_")
    
    # Select the columns we want
    pathways_table = pathways_table[["stId", "name", "entities_ratio", "entities_pValue", "entities_fdr", "entities_found", "entities_total"]]

    # If requested, filter out EHLD pathways
    if not include_high_level_diagrams:

        # Download a list of all diagrams with EHLD diagrams
        ehld_url = "https://reactome.org/download/current/ehld/svgsummary.txt"
        ehld_resp = requests.get(ehld_url)

        # Check that the response came back good
        if ehld_resp.status_code != requests.codes.ok:
            raise HttpResponseError(f"Checking whether pathways are high level pathways returned an HTTP status {ehld_resp.status_code}. The content returned from the request may be helpful:\n{ehld_resp.content.decode('utf-8')}")

        # Parse the response
        ehld_list = ehld_resp.content.decode("utf-8")
        ehld_list = ehld_list.split("\n")
        ehld_list = [pathway_id for pathway_id in ehld_list if pathway_id.startswith("R-HSA-")]

        has_ehld = pathways_table["stId"].isin(ehld_list)
        pathways_table = pathways_table[~has_ehld]

        # Make the index look normal
        pathways_table = pathways_table.reset_index(drop=True)

    return analysis_token, pathways_table



import pytest
import cptac
from cptac.exceptions import DataSourceNotFoundError

def test_error_handling():
    """Test the error handling for erroneous function calls"""
    brca = cptac.Brca()

    # Test a non-existent method
    with pytest.raises(AttributeError):
        brca.get_nonexistent_data()

    # Test with incorrect datatype
    with pytest.raises(DataSourceNotFoundError):
        brca.get_proteomics('invalid_datatype')

    # Test with incorrect source
    with pytest.raises(DataSourceNotFoundError):
        brca.get_dataframe("proteomics", source='nonexistent_source')

    # Test non-existent attribute
    with pytest.raises(AttributeError):
        brca.nonexistent_attribute